# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Alert TI enrichment - provides enrichment of alerts with threat intelligence."""
import json
from typing import Any, Dict, Iterable, Optional

import pandas as pd
from IPython.display import display
from msticpy.common.timespan import TimeSpan
from msticpy.common.utility import md
from tqdm.notebook import tqdm

# pylint: disable=ungrouped-imports
try:
    from msticpy.nbwidgets import SelectAlert
    from msticpy.vis.foliummap import FoliumMap, get_center_ip_entities
    from msticpy.vis.nbdisplay import format_alert
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools.foliummap import FoliumMap, get_center_ip_entities
    from msticpy.nbtools.nbdisplay import format_alert
    from msticpy.nbtools.nbwidgets import SelectAlert

from msticpy.nbtools.security_alert import SecurityAlert

from ...._version import VERSION
from ....common import (
    MsticnbDataProviderError,
    MsticnbMissingParameterError,
    nb_data_wait,
    nb_print,
    set_text,
)
from ....nb_metadata import read_mod_metadata
from ....nblib.iptools import convert_to_ip_entities
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Pete Bryan"


_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods
# Rename this class
class TIEnrichResult(NotebookletResult):
    """
    Template Results.

    Attributes
    ----------
    enriched_results : pd.DataFrame
        Alerts with additional TI enrichment
    picker : SelectAlert
        Alert picker

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[, optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.description: str = "Enriched Alerts"
        self.enriched_results: Optional[pd.DataFrame] = None
        self.picker: Optional[SelectAlert] = None

    @property
    def alert_selector(self):
        """Return the alert picker."""
        return self.picker

    @property
    def alert_picker(self):
        """Return the alert picker."""
        return self.picker


# pylint: enable=too-few-public-methods
# pylint: disable=too-many-branches
class EnrichAlerts(Notebooklet):
    """
    Alert Enrichment Notebooklet Class.

    Enriches Azure Sentinel alerts with TI data.

    """

    metadata = _CLS_METADATA

    @set_text(docs=_CELL_DOCS, key="run")
    def run(
        self,
        value: Optional[str] = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> TIEnrichResult:
        """
        Return an enriched set of Alerts.

        Parameters
        ----------
        timespan : TimeSpan
            Timespan for queries
        options : Optional[Iterable[str]], optional
            List of options to use, by default None.
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.
        value: Optional[str], optional
            If you want to filter Alerts based on a specific entity specify
            it as a string.
        data: Optional[pd.DataFrame], optional
            If you have alerts in a DataFrame you can pass them rather than
            having the notebooklet query alerts.

        Returns
        -------
        TIEnrichResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        MsticnbDataProviderError
            If data is not avaliable

        """
        # This line use logic in the superclass to populate options
        # (including default options) into this class.
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not timespan and data is None:
            raise MsticnbMissingParameterError("timespan.")

        # If data is not provided, query Sentinel to get it.
        if data is None:
            nb_print("Collecting alerts")
            if value is not None:
                data = _get_all_alerts(self.query_provider, timespan, value)
            else:
                data = _get_all_alerts(self.query_provider, timespan)

        # Create a result class
        # Add description to results for context
        self._last_result = TIEnrichResult(
            description=f"""Enriched alerts with the filter of {value}""",
            notebooklet=self,
            timespan=timespan,
        )

        # Establish TI providers
        if "tilookup" in self.data_providers.providers:
            ti_prov = self.data_providers.providers["tilookup"]
        else:
            raise MsticnbDataProviderError("No TI providers avaliable")

        if not isinstance(data, pd.DataFrame) or data.empty:
            raise MsticnbDataProviderError("No alerts avaliable")

        data["Entities"] = data["Entities"].apply(_entity_load)
        tqdm.pandas(desc="TI lookup progress")
        ti_sec = False
        if "secondary" in self.options:
            ti_sec = True
        md(
            """Alerts enriched with threat intelligence -
                 TI Risk is the the hightest score provided by any of
                 the configured providers."""
        )
        data["TI Risk"] = data.progress_apply(
            lambda row: _lookup(row, ti_prov, secondary=ti_sec), axis=1
        )
        if not self.silent:
            display(
                data[
                    ["StartTimeUtc", "AlertName", "Severity", "TI Risk", "Description"]
                ]
                .sort_values(by=["StartTimeUtc"])
                .style.map(_color_cells)
                .hide_index()
            )
        if "details" in self.options:
            geo_lookup = self.get_provider("geolitelookup") or self.get_provider(
                "ipstacklookup"
            )
            self._last_result.picker = _alert_picker(
                data,
                ti_prov,
                secondary=ti_sec,
                silent=self.silent,
                geo_lookup=geo_lookup,
            )
        self._last_result.enriched_results = data

        return self._last_result


# %%
# Display Alert Picker
@set_text(docs=_CELL_DOCS, key="select_alert")
def _alert_picker(data, ti_prov, secondary, silent: bool, geo_lookup: Any = None):
    ti_provs = "primary"
    if secondary is True:
        ti_provs = "all"

    def show_full_alert(selected_alert):
        global security_alert  # pylint: disable=global-variable-undefined, invalid-name
        output = []
        security_alert = SecurityAlert(selected_alert)
        output.append(format_alert(security_alert, show_entities=True))
        ioc_list = []
        if security_alert["Entities"] is not None:
            for entity in security_alert["Entities"]:
                if entity["Type"] in ("ipaddress", "ip"):
                    ioc_list.append(entity["Address"])
                elif entity["Type"] == "url":
                    ioc_list.append(entity["Url"])
            if ioc_list:
                ti_data = ti_prov.lookup_iocs(data=ioc_list, prov_scope=ti_provs)
                output.append(
                    ti_data[
                        ["Ioc", "IocType", "Provider", "Result", "Severity", "Details"]
                    ]
                    .reset_index()
                    .style.map(_color_cells)
                    .hide_index()
                )
                ti_ips = ti_data[ti_data["IocType"] == "ipv4"]
                # If we have IP entities try and plot these on a map
                if not ti_ips.empty:
                    ip_ents = [
                        convert_to_ip_entities(i, geo_lookup=geo_lookup)
                        for i in ti_ips["Ioc"].unique()
                    ]
                    ip_ents = [
                        ip_ent for ip_ent_list in ip_ents for ip_ent in ip_ent_list
                    ]
                    center = get_center_ip_entities(ip_ents)
                    ip_map = FoliumMap(location=center, zoom_start=4)
                    ip_map.add_ip_cluster(ip_ents, color="red")
                    output.append(ip_map)
                else:
                    output.append("")
            else:
                output.append("No IoCs")
        else:
            output.append("No Entities with IoCs")
        return output

    alert_select = SelectAlert(
        alerts=data,
        action=show_full_alert,
        columns=[
            "StartTimeUtc",
            "AlertName",
            "CompromisedEntity",
            "TI Risk",
            "SystemAlertId",
        ],
    )
    if silent is not True:
        alert_select.display()

    return alert_select


# pylint: disable=too-many-branches


# %%
# Get Alerts
def _get_all_alerts(qry_prov, timespan, filter_item=None):
    nb_data_wait("Alerts")
    if filter_item is not None:
        return qry_prov.SecurityAlert.list_alerts(
            start=timespan.start,
            end=timespan.end,
            add_query_items=f'| where Entities contains "{filter_item}"',
        )
    return qry_prov.SecurityAlert.list_alerts(start=timespan.start, end=timespan.end)


# Extract packed entity details
def _entity_load(entity):
    try:
        return json.loads(entity)
    except json.JSONDecodeError:
        return None


# Lookup details against TI
def _lookup(row, ti_prov, secondary: bool = False):
    sev = []
    prov_scope = "primary"
    if secondary:
        prov_scope = "all"
    if row["Entities"] is not None:
        for entity in row["Entities"]:
            try:
                if entity["Type"] in ("ipaddress", "ip"):
                    resp = ti_prov.lookup_ioc(entity["Address"], prov_scope=prov_scope)
                elif entity["Type"] == "url":
                    resp = ti_prov.lookup_ioc(entity["Url"], prov_scope=prov_scope)
                else:
                    resp = None
                if resp:
                    for response in resp[1]:
                        sev.append(response[1].severity)
            except KeyError:
                pass

    return _sev_score(sev)


# Color code cells based on TI results
def _color_cells(val):
    color_chart = {"high": "Red", "warning": "Orange", "information": "Green"}
    if isinstance(val, str) and val.casefold() in color_chart:
        color = color_chart[val.casefold()]
    else:
        color = "none"
    return f"background-color: {color}"


def _sev_score(sev):
    if "high" in sev:
        return "High"
    if "warning" in sev:
        return "Warning"
    if "information" in sev:
        return "Information"
    return "None"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Alert utility functions."""
try:
    from msticpy import nbwidgets
    from msticpy.vis import nbdisplay

    if not hasattr(nbwidgets, "SelectItem"):
        raise ImportError("Invalid nbwidgets")
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools import nbdisplay, nbwidgets

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def browse_alerts(nb_result, alert_attr="related_alerts") -> nbwidgets.SelectAlert:
    """Return alert browser/viewer."""
    if nb_result is None or not hasattr(nb_result, alert_attr):
        return None
    rel_alerts = getattr(nb_result, alert_attr, None)
    if rel_alerts is None or rel_alerts.empty:
        return None

    if "CompromisedEntity" not in rel_alerts:
        rel_alerts["CompromisedEntity"] = "n/a"
    if "StartTimeUtc" not in rel_alerts:
        rel_alerts["StartTimeUtc"] = rel_alerts["TimeGenerated"]
    return nbwidgets.SelectAlert(alerts=rel_alerts, action=nbdisplay.format_alert)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""common test class."""
import io
import warnings
from contextlib import redirect_stdout

import pytest
import pytest_check as check

from msticnb import init, options
from msticnb.common import add_result, nb_data_wait, nb_debug, nb_print
from msticnb.options import get_opt, set_opt

from .nb_test import TstNBSummary

# pylint: disable=too-many-statements


def test_print_methods():
    """Test method."""
    set_opt("verbose", True)
    f_stream = io.StringIO()
    with redirect_stdout(f_stream):
        nb_print("status")
        nb_data_wait("table1")
    check.is_in("status", str(f_stream.getvalue()))
    check.is_in("Getting data from table1", str(f_stream.getvalue()))

    set_opt("verbose", False)
    f_stream = io.StringIO()
    with redirect_stdout(f_stream):
        nb_print("status")
    check.is_not_in("status", str(f_stream.getvalue()))
    check.is_not_in("Getting data from table1", str(f_stream.getvalue()))

    set_opt("debug", True)
    f_stream = io.StringIO()
    with redirect_stdout(f_stream):
        nb_debug("debug", "debugmssg", "val", 1, "result", True)
    check.is_in("debug", str(f_stream.getvalue()))
    check.is_in("debugmssg", str(f_stream.getvalue()))
    check.is_in("val", str(f_stream.getvalue()))
    check.is_in("1", str(f_stream.getvalue()))
    check.is_in("result", str(f_stream.getvalue()))
    check.is_in("True", str(f_stream.getvalue()))


def test_add_result_decorator():
    """Test method."""

    # pylint: disable=too-few-public-methods
    class _TestClass:
        prop1 = None
        prop2 = None

    test_obj = _TestClass()

    @add_result(result=test_obj, attr_name=["prop1", "prop2"])
    def test_func():
        return "result1", 10

    test_func()
    check.equal("result1", test_obj.prop1)
    check.equal(10, test_obj.prop2)


def test_options():
    """Test method."""
    set_opt("verbose", True)
    f_stream = io.StringIO()
    with redirect_stdout(f_stream):
        options.current()
    check.is_in("verbose: True", str(f_stream.getvalue()))

    f_stream = io.StringIO()
    with redirect_stdout(f_stream):
        options.show()
    check.is_in(
        "verbose (default=True): Show progress messages.", str(f_stream.getvalue())
    )

    with pytest.raises(KeyError):
        get_opt("no_option")

    with pytest.raises(KeyError):
        set_opt("no_option", "value")

    # This will work since bool(10) == True
    set_opt("verbose", 10)


def _capture_nb_run_output(test_nb, **kwargs):
    f_stream = io.StringIO()
    with redirect_stdout(f_stream):
        test_nb.run(**kwargs)
    return str(f_stream.getvalue())


def test_silent_option():
    """Test operation of 'silent' option."""
    warnings.filterwarnings(action="ignore", category=UserWarning)
    init(query_provider="LocalData", providers=[])
    test_nb = TstNBSummary()

    output = _capture_nb_run_output(test_nb)
    check.is_true(output)

    # Silent option to run
    output = _capture_nb_run_output(test_nb, silent=True)
    check.is_false(output)
    check.is_true(get_opt("silent"))

    # Silent option to init
    test_nb = TstNBSummary(silent=True)
    check.is_true(test_nb.silent)
    output = _capture_nb_run_output(test_nb)
    check.is_false(output)

    # But overridable on run
    output = _capture_nb_run_output(test_nb, silent=False)
    check.is_true(output)
    check.is_false(get_opt("silent"))

    # Silent global option
    set_opt("silent", True)
    test_nb = TstNBSummary()
    output = _capture_nb_run_output(test_nb)
    check.is_false(output)

    # But overridable on run
    output = _capture_nb_run_output(test_nb, silent=False)
    check.is_true(output)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet for Host Summary."""
from functools import lru_cache
from typing import Any, Dict, Iterable, Optional

import msticpy
import pandas as pd
from bokeh.models import LayoutDOM
from IPython.display import display

try:
    from msticpy.context.ip_utils import ip_whois as get_whois_info
    from msticpy.vis.foliummap import FoliumMap
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools.foliummap import FoliumMap
    from msticpy.sectools.ip_utils import get_whois_info

from msticpy.common.timespan import TimeSpan
from msticpy.common.utility import md
from msticpy.datamodel.entities import Host

from ...._version import VERSION
from ....common import (
    MsticnbDataProviderError,
    MsticnbMissingParameterError,
    nb_data_wait,
    nb_markdown,
    set_text,
)
from ....nb_metadata import read_mod_metadata, update_class_doc
from ....nblib.ti import get_ti_results
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Pete Bryan"


_MSTICPY_VER = msticpy.__version__
_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods
class HostNetworkSummaryResult(NotebookletResult):
    """Host Network Summary Results."""

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """Create new Notebooklet result instance."""
        super().__init__(description, timespan, notebooklet)
        self.flows: Optional[pd.DataFrame] = None
        self.flow_matrix: Optional[LayoutDOM] = None
        self.flow_whois: Optional[pd.DataFrame] = None
        self.flow_map: Optional[FoliumMap] = None
        self.flow_ti: Optional[pd.DataFrame] = None


# pylint: disable=too-few-public-methods
class HostNetworkSummary(Notebooklet):
    """
    HostSummary Notebooklet class.

    Queries and displays information about a host including:

    - IP address assignment
    - Related alerts
    - Related hunting/investigation bookmarks
    - Azure subscription/resource data.

    """

    metadata = _CLS_METADATA
    __doc__ = update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    def __init__(self, *args, **kwargs):
        """Initialize the Host Network Summary notebooklet."""
        super().__init__(*args, **kwargs)

    # pylint: disable=too-many-branches, too-many-locals
    @set_text(docs=_CELL_DOCS, key="run")  # noqa: MC0001
    def run(  # noqa:MC0001, C901
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> HostNetworkSummaryResult:
        """
        Return host summary data.

        Parameters
        ----------
        value : str
            Host entity
        data : Optional[pd.DataFrame], optional
            Not used, by default None
        timespan : TimeSpan
            Timespan over which operations such as queries will be
            performed, by default None.
            This can be a TimeStamp object or another object that
            has valid `start`, `end`, or `period` attributes.
        options : Optional[Iterable[str]], optional
            List of options to use, by default None
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Other Parameters
        ----------------
        start : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.
        end : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.

        Returns
        -------
        HostSummaryResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not value:
            raise MsticnbMissingParameterError("value")
        if not timespan:
            raise MsticnbMissingParameterError("timespan.")

        self.timespan = timespan

        # pylint: disable=attribute-defined-outside-init
        result = HostNetworkSummaryResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )

        if isinstance(value, Host):
            host_name = value.HostName
            ip_addr = value.IpAddress.Address if "IpAddress" in value else None
        elif isinstance(value, tuple):
            host_name, ip_addr = value
        else:
            raise ValueError(
                "Could not determine host name or IP address from value parameter."
                "Please supply host and ip address in the form of a Host entity",
                "or tuple of (host_name, ip_address).",
            )

        if not host_name and not ip_addr:
            md(f"Could not obtain unique host name from {value}. Aborting.")
            self._last_result = result
            return self._last_result

        result.flows = _get_host_flows(
            host_name=host_name,
            ip_addr=ip_addr,
            qry_prov=self.query_provider,
            timespan=self.timespan,
        )
        if result.flows is None:
            nb_markdown("No network flow data found.")
            self._last_result = result
            return self._last_result

        remote_ip_col = "RemoteIP"
        local_ip_col = "LocalIP"
        if "SourceIP" in result.flows.columns:
            remote_ip_col = "DestinationIP"
            local_ip_col = "SourceIP"
        if not result.flows.empty:
            result.flow_matrix = result.flows.mp_plot.matrix(
                x=remote_ip_col, y=local_ip_col, title="IP Interaction", sort="asc"
            )

        flows_remote_ips = result.flows[[remote_ip_col]].drop_duplicates()
        flows_source_ips = (
            result.flows[[local_ip_col]]
            .drop_duplicates()
            .rename(columns={local_ip_col: remote_ip_col})
        )
        flows_all = pd.concat([flows_remote_ips, flows_source_ips]).drop_duplicates()
        if "ti" in self.options:
            if "tilookup" in self.data_providers.providers:
                ti_prov = self.data_providers.providers["tilookup"]
            else:
                raise MsticnbDataProviderError("No TI providers available")
            ti_results, ti_results_merged = get_ti_results(
                ti_prov, flows_all, remote_ip_col
            )
            if isinstance(ti_results, pd.DataFrame) and not ti_results.empty:
                result.flow_ti = ti_results_merged

        if (
            "map" in self.options
            and isinstance(flows_all, pd.DataFrame)
            and not flows_all.empty
        ):
            result.flow_map = flows_all.mp_plot.folium_map(ip_column=remote_ip_col)

        if "whois" in self.options:
            result.flow_whois = _get_whois_data(flows_all, col=remote_ip_col)

        self._last_result = result

        if not self.silent:
            self._display_results()

        return self._last_result

    @set_text(docs=_CELL_DOCS, key="flows")
    def _display_flows(self):
        if self.check_valid_result_data("flow_whois", silent=True):
            display(self._last_result.flow_whois)
        elif self.check_valid_result_data("flows", silent=True):
            display(self._last_result.flows)
        else:
            nb_markdown("No network flow data found.")

    @set_text(docs=_CELL_DOCS, key="ti")
    def _display_ti_results(self):
        if self.check_valid_result_data("flow_ti", silent=True):
            display(self._last_result.flow_ti)
        else:
            nb_markdown("No Threat Intelligence results found.")

    @set_text(docs=_CELL_DOCS, key="map")
    def _display_map(self):
        if (
            self.check_valid_result_data("flows", silent=True)
            and self._last_result.flow_map
        ):
            display(self._last_result.flow_map)

    @set_text(docs=_CELL_DOCS, key="matrix")
    def _display_matrix(self):
        if self._last_result.flow_matrix:
            display(self._last_result.flow_matrix)

    def _display_results(self):
        self._display_flows()
        self._display_ti_results()
        self._display_map()
        self._display_matrix()


@lru_cache()
def _get_host_flows(host_name, ip_addr, qry_prov, timespan) -> Optional[pd.DataFrame]:
    if host_name:
        nb_data_wait("Host flow events")
        host_flows = qry_prov.MDE.host_connections(timespan, host_name=host_name)
        host_flows["SourceIP"] = host_flows["LocalIP"]
        host_flows["DestinationIP"] = host_flows["RemoteIP"]
        host_flows_csl = qry_prov.Network.host_network_connections_csl(
            timespan, host_name=host_name
        )
    elif ip_addr:
        nb_data_wait("Host flow events")
        host_flows = qry_prov.Network.list_azure_network_flows_by_ip(
            timespan, ip_address_list=[ip_addr]
        )
        host_flows_csl = qry_prov.Network.ip_network_connections_csl(
            timespan, ip=ip_addr
        )
    else:
        return None
    return pd.concat([host_flows, host_flows_csl], sort=False)


def _get_whois_data(data, col) -> pd.DataFrame:
    if not data.empty:
        if _MSTICPY_VER < "2.13.0":
            data["ASN"] = data.apply(lambda x: get_whois_info(x[col]), axis=1)
        else:
            data["ASN"] = data.apply(
                lambda x: get_whois_info(x[col], True).name, axis=1
            )
            data["ASNProperties"] = data.apply(
                lambda x: get_whois_info(x[col], True).properties, axis=1
            )
    return data



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""data_providers test class."""
import sys

import pytest_check as check
from msticpy.data import QueryProvider

try:
    from msticpy.context import TILookup
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.sectools import TILookup

from msticnb import data_providers

from .unit_test_lib import GeoIPLiteMock

# pylint: disable=no-member


def test_init_data_providers(monkeypatch):
    """Test creating DataProviders instance."""
    monkeypatch.setattr(data_providers, "GeoLiteLookup", GeoIPLiteMock)
    dprov = data_providers.DataProviders(query_provider="LocalData")

    check.is_not_none(dprov)
    check.equal(dprov, data_providers.DataProviders.current())

    check.is_in("LocalData", dprov.providers)
    check.is_in("geolitelookup", dprov.providers)
    check.is_in("tilookup", dprov.providers)
    check.is_instance(dprov.providers["LocalData"], QueryProvider)
    check.is_instance(dprov.providers["geolitelookup"], GeoIPLiteMock)
    check.is_instance(dprov.providers["tilookup"], TILookup)


def test_new_init_data_providers(monkeypatch):
    """Test creating new provider with new provider list."""
    monkeypatch.setattr(data_providers, "GeoLiteLookup", GeoIPLiteMock)

    data_providers.init(query_provider="LocalData", providers=[])
    dprov = data_providers.DataProviders.current()
    data_providers.init(query_provider="LocalData", providers=[])
    dprov2 = data_providers.DataProviders.current()
    check.equal(dprov2, dprov)

    # specify provider
    dprov = data_providers.DataProviders(query_provider="LocalData")
    data_providers.init(query_provider="LocalData", providers=["tilookup"])
    msticnb = sys.modules["msticnb"]
    dprov2 = data_providers.DataProviders.current()
    pkg_providers = getattr(msticnb, "data_providers")
    check.not_equal(dprov2, dprov)
    check.is_in("LocalData", dprov2.providers)
    check.is_in("tilookup", dprov2.providers)
    check.is_not_in("geolitelookup", dprov2.providers)
    check.is_not_in("ipstacklookup", dprov2.providers)
    check.is_in("LocalData", pkg_providers)
    check.is_in("tilookup", pkg_providers)
    check.is_not_in("geolitelookup", pkg_providers)
    check.is_not_in("ipstacklookup", pkg_providers)

    check.is_instance(dprov2.providers["tilookup"], TILookup)


def test_add_sub_data_providers(monkeypatch):
    """Test intializing adding and subtracting providers."""
    monkeypatch.setattr(data_providers, "GeoLiteLookup", GeoIPLiteMock)

    dprov = data_providers.DataProviders(query_provider="LocalData")
    data_providers.init(query_provider="LocalData", providers=["tilookup"])
    msticnb = sys.modules["msticnb"]
    dprov2 = data_providers.DataProviders.current()

    # Add and remove a provider from defaults
    data_providers.init(
        query_provider="LocalData", providers=["+ipstacklookup", "-geolitelookup"]
    )

    dprov3 = data_providers.DataProviders.current()
    pkg_providers = getattr(msticnb, "data_providers")
    check.not_equal(dprov3, dprov)
    check.not_equal(dprov3, dprov2)
    check.is_in("ipstacklookup", dprov3.providers)
    check.is_not_in("geolitelookup", dprov3.providers)
    check.is_in("tilookup", dprov3.providers)
    check.is_in("ipstacklookup", pkg_providers)
    check.is_not_in("geolitelookup", pkg_providers)
    check.is_in("tilookup", pkg_providers)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""NB metadata test class."""
import pytest_check as check

from msticnb import data_providers, init, nblts
from msticnb.nb.azsent.host import host_summary
from msticnb.nb_metadata import NBMetadata, read_mod_metadata

from .unit_test_lib import GeoIPLiteMock


def test_read_metadata():
    """Tests reading metadata yaml file."""
    nb_md, docs = read_mod_metadata(host_summary.__file__, host_summary.__name__)
    check.is_instance(nb_md, NBMetadata)
    check.is_instance(docs, dict)

    opts = nb_md.get_options("all")
    check.is_in("heartbeat", [opt[0] for opt in opts])
    check.is_in("alerts", [opt[0] for opt in opts])

    for item in ("Default Options", "alerts", "azure_api"):
        check.is_in(item, nb_md.options_doc)

    for item in ("Default Options", "alerts", "azure_api"):
        check.is_in(item, nb_md.options_doc)


# pylint: disable=protected-access
def test_class_metadata(monkeypatch):
    """Test class correctly loads yaml metadata."""
    init()
    monkeypatch.setattr(data_providers, "GeoLiteLookup", GeoIPLiteMock)
    if "azuredata" in nblts.azsent.host.HostSummary.metadata.req_providers:
        nblts.azsent.host.HostSummary.metadata.req_providers.remove("azuredata")
    data_providers.init(
        query_provider="LocalData", providers=["tilookup", "geolitelookip"]
    )
    host_nb = host_summary.HostSummary()

    check.is_true(hasattr(host_summary, "_CLS_METADATA"))
    check.is_instance(host_summary._CLS_METADATA, NBMetadata)
    check.is_true(hasattr(host_summary, "_CELL_DOCS"))
    check.is_instance(host_summary._CELL_DOCS, dict)

    check.is_true(hasattr(host_nb, "metadata"))
    check.is_instance(host_nb.metadata, NBMetadata)
    check.equal(host_nb.metadata.mod_name, host_summary.__name__)
    check.equal(host_nb.description(), "Host summary")
    check.equal(host_nb.name(), "HostSummary")
    check.is_in("host", host_nb.entity_types())
    check.is_in("host", host_nb.keywords())

    check.is_in("heartbeat", host_nb.default_options())
    check.is_in("alerts", host_nb.default_options())

    check.is_in("alerts", host_nb.all_options())

    for item in ("Default Options", "alerts", "azure_api"):
        check.is_in(item, host_nb.list_options())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet base classes."""
import inspect
import re
import warnings
from abc import ABC, abstractmethod
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd
from IPython.core.getipython import get_ipython
from IPython.display import HTML, display
from msticpy.common.timespan import TimeSpan
from tqdm.auto import tqdm

from ._version import VERSION
from .common import MsticnbDataProviderError, MsticnbError
from .data_providers import DataProviders
from .nb_metadata import NBMetadata, read_mod_metadata
from .notebooklet_result import NotebookletResult
from .options import get_opt, set_opt

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-public-methods
class Notebooklet(ABC):
    """Base class for Notebooklets."""

    metadata: NBMetadata = NBMetadata(
        name="Notebooklet", description="Base class", default_options=[]
    )
    module_path: Union[str, Path] = ""

    def __init__(self, data_providers: Optional[DataProviders] = None, **kwargs):
        """
        Initialize a new instance of the notebooklet class.

        Parameters
        ----------
        data_providers : DataProviders, Optional
            Optional DataProviders instance to query data.
            Most classes require this.

        Raises
        ------
        MsticnbDataProviderError
            If DataProviders has not been initialized.
            If required providers are specified by the notebooklet
            but are not available.

        """
        self._kwargs = kwargs
        self.options: List[str] = self.default_options()
        self._set_tqdm_notebook(get_opt("verbose"))
        self._last_result: Any = None
        self.timespan = TimeSpan(period="1d")
        self._inst_default_silent: Optional[bool] = kwargs.get("silent")
        self._current_run_silent: Optional[bool] = None
        set_opt("temp_silent", self.silent)

        # update "run" function documentation on first run
        self._add_run_doc_options()

        # Check required data providers are loaded.
        # pylint: disable=no-member
        self.data_providers = data_providers or DataProviders.current()  # type: ignore
        # pylint: enable=no-member
        self._check_nb_providers(**kwargs)

    def _check_nb_providers(self, **kwargs):
        """Check that providers required for notebooklet are available."""
        if not self.data_providers:
            raise MsticnbDataProviderError(
                "No current DataProviders instance was found.",
                "Please create an instance of msticnb.",
            )
        self.query_provider = self.data_providers.query_provider
        missing_provs, unknown_provs = self.data_providers.has_required_providers(
            self.metadata.req_providers
        )
        prov_add_errs: List[Exception] = []
        if missing_provs:
            # Try to add any providers that the notebooklet needs.
            add_providers = missing_provs.copy()
            for missing_prov in add_providers:
                try:
                    self.data_providers.add_provider(provider=missing_prov, **kwargs)
                    missing_provs.remove(missing_prov)
                    if missing_prov in unknown_provs:
                        unknown_provs.remove(missing_prov)
                except MsticnbDataProviderError as err:
                    prov_add_errs.append(err)

        if missing_provs:
            missing_mssg = [prov.replace("|", " or ") for prov in missing_provs]
            raise MsticnbDataProviderError(
                f"Required data provider(s) {', '.join(missing_mssg)} not loaded.",
                f"Class {self.__class__.__name__}",
                *prov_add_errs,
            )
        if unknown_provs:
            warnings.warn(
                f"Unknown provider(s) {', '.join(unknown_provs)} in req_providers list."
                + f"Class {self.__class__.__name__}"
            )

    def _add_run_doc_options(self):
        """Add options documentation to run function."""
        if "Default Options" in self.__class__.run.__doc__:
            return
        options_doc = (f"    {line}" for line in self.metadata.options_doc.split("\n"))
        self.__class__.run.__doc__ = (self.__class__.run.__doc__ or "") + "\n".join(
            options_doc
        )

    @abstractmethod
    def run(
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> NotebookletResult:
        """
        Notebooklet abstract base class.

        Parameters
        ----------
        value : Any, optional
            value to process, by default None
        data : Optional[pd.DataFrame], optional
            Input data to process, by default None
        timespan : Optional[TimeSpan, Any], optional
            Timespan over which operations such as queries will be
            performed, by default None.
            This can be a TimeStamp object or another object that
            has valid `start`, `end`, or `period` attributes.
        options :Optional[Iterable[str]], optional
            List of options to use, by default None
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            Options prefixed with "-" will be removed from the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Other Parameters
        ----------------
        start : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.
        end : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.

        Returns
        -------
        NotebookletResult
            Result class from the notebooklet

        See Also
        --------
            TimeSpan

        """
        self._current_run_silent = kwargs.get("silent")
        set_opt("temp_silent", self.silent)
        if not options:
            self.options = self.default_options()
        else:
            def_options = self.default_options()
            add_options = {opt[1:] for opt in options if opt.startswith("+")}
            sub_options = {opt[1:] for opt in options if opt.startswith("-")}
            std_options = {opt for opt in options if opt[0] not in ("+", "-")}
            if std_options and (add_options or sub_options):
                raise MsticnbError(
                    "Option list must be either a list of options to use",
                    "or options to add/remove from the default set.",
                    "You cannot mix these.",
                )
            invalid_opts = (sub_options | add_options | std_options) - set(
                self.all_options()
            )
            if invalid_opts:
                print(f"Invalid options {list(invalid_opts)} ignored.")
            if sub_options:
                self.options = list(set(def_options) - sub_options)
            if add_options:
                self.options = list(set(def_options) | add_options)
            if not (add_options or sub_options):
                self.options = list(options)
        self._set_tqdm_notebook(get_opt("verbose"))
        if timespan:
            self.timespan = TimeSpan(timespan=timespan)
        elif "start" in kwargs and "end" in kwargs:
            self.timespan = TimeSpan(start=kwargs.get("start"), end=kwargs.get("end"))
        return NotebookletResult(notebooklet=self)

    def get_pivot_run(self, get_timespan: Callable[[], TimeSpan]):
        """Return Pivot-wrappable run function."""

        @wraps
        def pivot_run(*args, **kwargs):
            result = self.run(*args, timespan=get_timespan(), **kwargs)
            setattr(result, "notebooklet", self)
            return result

        return pivot_run

    def get_provider(self, provider_name: str):
        """
        Return data provider for the specified name.

        Parameters
        ----------
        provider_name : str
            Name of the provider

        Returns
        -------
        Any
            Provider instance.

        Raises
        ------
        MsticnbDataProviderError
            If provider is not found.

        """
        if provider_name not in self.data_providers.providers:
            raise MsticnbDataProviderError(
                f"Data provider {provider_name} not found.",
                "Please check that you have specified the required providers",
            )
        return self.data_providers.providers.get(provider_name)

    @property
    def silent(self) -> Optional[bool]:
        """
        Get the current instance setting for silent running.

        Returns
        -------
        Optional[bool]
            Silent running is enabled.

        """
        if self._current_run_silent is not None:
            return self._current_run_silent
        if self._inst_default_silent is not None:
            return self._inst_default_silent
        return None

    @silent.setter
    def silent(self, value: bool):
        """
        Set the current instance setting for silent running.

        Parameters
        ----------
        value : bool
            True to enable silent, False to disable.

        """
        self._inst_default_silent = value

    @property
    def result(self) -> Optional[NotebookletResult]:
        """
        Return result of the most recent notebooklet run.

        Returns
        -------
        Optional[NotebookletResult]
            Notebooklet result class or None if nothing has
            been run.

        """
        return self._last_result

    @classmethod
    def name(cls) -> str:
        """
        Return name of the Notebooklet.

        Returns
        -------
        str
            Name

        """
        return cls.metadata.name

    @classmethod
    def description(cls) -> str:
        """
        Return description of the Notebooklet.

        Returns
        -------
        str
            Description

        """
        return cls.metadata.description

    @classmethod
    def all_options(cls) -> List[str]:
        """
        Return supported options for Notebooklet run function.

        Returns
        -------
        List[str]
            Supported options.

        """
        opts = cls.metadata.get_options("all")
        return [opt[0] for opt in opts]

    @classmethod
    def default_options(cls) -> List[str]:
        """
        Return default options for Notebooklet run function.

        Returns
        -------
        List[str]
            Supported options.

        """
        opts = cls.metadata.get_options("default")
        return [opt[0] for opt in opts]

    @classmethod
    def list_options(cls) -> str:
        """
        Return options document for Notebooklet run function.

        Returns
        -------
        List[str]
            Supported options.

        """
        return cls.metadata.options_doc

    @classmethod
    def print_options(cls):
        """Print options for Notebooklet run function."""
        print(cls.metadata.options_doc)

    @classmethod
    def keywords(cls) -> List[str]:
        """
        Return search keywords for Notebooklet.

        Returns
        -------
        List[str]
            Keywords

        """
        return cls.metadata.keywords

    @classmethod
    def entity_types(cls) -> List[str]:
        """
        Entity types supported by the notebooklet.

        Returns
        -------
        List[str]
            Entity names

        """
        return cls.metadata.entity_types

    @classmethod
    def get_settings(cls, print_settings=True) -> Optional[str]:
        """
        Print or return metadata for class.

        Parameters
        ----------
        print_settings : bool, optional
            Print to standard, by default True
            or return the str formatted content.

        Returns
        -------
        Optional[str]
            If `print_settings` is True, returns None.
            If False, returns LF-delimited string of metadata settings.

        Notes
        -----
            Use `metadata` attribute to retrieve the metadata directly.

        """
        if print_settings:
            print(cls.metadata)
            return None
        return str(cls.metadata)

    @classmethod
    def match_terms(cls, search_terms: str) -> Tuple[bool, int]:
        """
        Search class definition for `search_terms`.

        Parameters
        ----------
        search_terms : str
            One or more search terms, separated by spaces
            or commas.
            Terms can be simple strings or regular expressions.

        Returns
        -------
        Tuple[bool, int]
            Returns a tuple of bool (True if all terms match)
            and int (count of matched terms)

        """
        search_text = " ".join(cls.metadata.search_terms)
        search_text += cls.__doc__ or ""
        terms = [
            subterm for term in search_terms.split(",") for subterm in term.split()
        ]
        match_count = sum(
            1 for term in terms if re.search(term, search_text, re.IGNORECASE)
        )

        return match_count == len(terms), match_count

    @staticmethod
    def _set_tqdm_notebook(verbose=False):
        if verbose:
            tqdm.pandas()

    def _get_timespan(self, timespan=None, **kwargs):
        if timespan:
            if isinstance(timespan, TimeSpan):
                self.timespan = timespan
            else:
                self.timespan = TimeSpan(timespan=timespan)
        elif "start" in kwargs and "end" in kwargs:
            self.timespan = TimeSpan(start=kwargs["start"], end=kwargs["end"])

    @classmethod
    def import_cell(cls):
        """Import the text of this module into a new cell."""
        if cls.module_path:
            with open(cls.module_path, "r", encoding="utf-8") as mod_file:
                mod_text = mod_file.read()
            if mod_text:
                # replace relative references with absolute paths
                mod_text = cls._update_mod_for_import(cls.module_path, mod_text)
                shell = get_ipython()
                shell.set_next_input(mod_text)

    @classmethod
    def _update_mod_for_import(cls, module_path, mod_text):
        mod_text = re.sub(r"\.{3,}", "msticnb.", mod_text)
        metadata, docs = read_mod_metadata(module_path, cls.__module__)
        metadata_repr = repr(metadata)
        metadata_repr = metadata_repr.replace("NBMetadata", "nb_metadata.NBMetadata")
        repl_text = (
            "_CLS_METADATA, _CELL_DOCS = "
            + "nb_metadata.read_mod_metadata(__file__, __name__)"
        )
        inline_text = f"_CELL_DOCS = {str(docs)}\n"
        inline_text = f"{inline_text}\n_CLS_METADATA = {metadata_repr}"
        return mod_text.replace(repl_text, inline_text)

    @classmethod
    def show_help(cls):
        """Display Documentation for class."""
        display(HTML(cls.get_help()))

    @classmethod
    def get_help(cls, fmt="html") -> str:
        """Return HTML document for class."""
        return cls._get_doc(fmt=fmt)

    @classmethod
    def _get_doc(cls, fmt):
        """Return documentation func. placeholder."""
        del fmt
        return "No documentation available."

    def check_valid_result_data(
        self, attrib: Optional[str] = None, silent: bool = False
    ) -> bool:
        """
        Check that the result is valid and `attrib` contains data.

        Parameters
        ----------
        attrib : str
            Name of the attribute to check, if None this function
            only checks for a valid _last_result.
        silent : bool
            If True, suppress output.

        Returns
        -------
        bool
            Returns True if valid data is available, else False.

        """
        if self._last_result is None:
            if not silent:
                print(
                    "No current result."
                    "Please use 'run()' to fetch the data before using this method."
                )
            return False
        if not attrib:
            return True
        data_obj = getattr(self._last_result, attrib)
        if data_obj is None or isinstance(data_obj, pd.DataFrame) and data_obj.empty:
            if not silent:
                print(f"No data is available for last_result.{attrib}.")
            return False
        return True

    def check_table_exists(self, table: str) -> bool:
        """
        Check to see if the table exists in the provider.

        Parameters
        ----------
        table : str
            Table name

        Returns
        -------
        bool
            True if the table exists, otherwise False.

        """
        if not self.query_provider:
            print(f"No query provider for table {table} is available.")
            return False
        if table not in self.query_provider.schema_tables:
            print(f"table {table} is not available.")
            return False
        return True

    def get_methods(self) -> Dict[str, Callable[[Any], Any]]:
        """Return methods available for this class."""
        meths = inspect.getmembers(self, inspect.ismethod)
        cls_selector = (
            f"bound method {self.__class__.__name__.rsplit('.', maxsplit=1)[0]}"
        )
        return {
            meth[0]: meth[1]
            for meth in meths
            if cls_selector in str(meth[1]) and not meth[0].startswith("_")
        }

    def list_methods(self) -> List[str]:
        """Return list of methods with descriptions."""
        methods = self.get_methods()
        method_desc: List[str] = []
        for name, method in methods.items():
            f_doc = inspect.getdoc(method)
            desc = f_doc.split("\n", maxsplit=1)[0] if f_doc else ""
            method_desc.append(f"{name} - '{desc}'")
        return method_desc



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Add notebooklets as pivot functions."""
from functools import wraps
from typing import Any, Callable, Dict

from msticpy.common.timespan import TimeSpan

try:
    from msticpy.init.pivot import Pivot, PivotRegistration
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.datamodel.pivot import Pivot, PivotRegistration

from ._version import VERSION
from .notebooklet import Notebooklet
from .read_modules import nblts

__version__ = VERSION
__author__ = "Ian Hellen"


_ENTITY_MAP = {
    "host": {"Host": "HostName"},
    "account": {"Account": "Name"},
    "ip_address": {"IpAddress": "Address"},
    "alert": {"Alert": "AlertType"},
}


def _to_py_name(name: str) -> str:
    func_name = "".join(char if char.islower() else f"_{char.lower()}" for char in name)
    return func_name.strip("_")


def add_pivot_funcs(pivot: Pivot = None, **kwargs):
    """
    Add notebooklet run functions as pivot methods.

    Parameters
    ----------
    pivot : Pivot, optional
        Pivot instance.
    kwargs : Dict[str, Any]
        Optional keyword arguments to pass to Pivot init.

    """
    if not pivot:
        piv_kwargs = {
            key: arg for key, arg in kwargs.items() if key in ("namespace", "providers")
        }
        pivot = Pivot.current() or Pivot(**piv_kwargs)
    for nb_name, nb_class in nblts.iter_classes():
        if not issubclass(nb_class, Notebooklet) or nb_name == "TemplateNB":
            continue
        if "value" not in nb_class.metadata.inputs:
            # This doesn't take a "value" input so can't use as a pivot
            continue
        nb_obj = nb_class()
        run_func = getattr(nb_obj, "run")
        pivot_timespan = pivot.get_timespan

        wrp_func = _wrap_run_func(run_func, pivot_timespan)
        func_new_name = _to_py_name(nb_name)
        entity_map: Dict[str, str] = {}
        for entity in nb_class.metadata.entity_types:
            if entity not in _ENTITY_MAP:
                continue
            entity_map.update(_ENTITY_MAP[entity])

        if not entity_map:
            continue

        piv_reg = PivotRegistration(
            input_type="value",
            entity_map=entity_map,
            func_new_name=func_new_name,
            src_func_name="run",
            can_iterate=False,
            func_input_value_arg="value",
            return_raw_output=True,
        )
        Pivot.add_pivot_function(func=wrp_func, pivot_reg=piv_reg, container="nblt")


def _wrap_run_func(func: Callable[[Any], Any], get_time_span: Callable[[], TimeSpan]):
    """Wrap function to inject timespan."""

    @wraps(func)
    def _wrapped_func(*args, **kwargs):
        time_span = get_time_span()
        kwargs.update({"timespan": time_span})
        return func(*args, **kwargs)

    return _wrapped_func



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Common definitions and classes."""
import functools
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import bokeh.io
import pandas as pd
from IPython import get_ipython
from IPython.display import HTML, display
from markdown import markdown
from msticpy import VERSION as MP_VERSION
from msticpy.common import utility as mp_utils
from packaging.version import parse

from ._version import VERSION
from .options import get_opt

__version__ = VERSION
__author__ = "Ian Hellen"


_IP_AVAILABLE = get_ipython() is not None


class NBContainer:
    """Container for Notebooklet classes."""

    def __len__(self):
        """Return number of items in the attribute collection."""
        return len(self.__dict__)

    def __iter__(self):
        """Return iterator over the attributes."""
        return iter(self.__dict__.items())

    def __repr__(self):
        """Return list of attributes."""
        obj_list = []
        for key, val in self.__dict__.items():
            if isinstance(val, NBContainer):
                obj_list.append(f"{key} [{repr(val)}]")
            else:
                obj_list.append(repr(val))
        return ", ".join(obj_list)

    def __str__(self):
        """Print a string representation of the object."""
        obj_str = ""
        for key, val in self.__dict__.items():
            if isinstance(val, NBContainer):
                obj_str += f"{key}\n"
                for line in str(val).split("\n"):
                    if line.strip():
                        obj_str += f"  {line}\n"
            else:
                obj_str += val.__name__ + " (Notebooklet)\n"
        return obj_str

    def iter_classes(self) -> Iterable[Tuple[str, Any]]:
        """Return iterator through all notebooklet classes."""
        for key, val in self.__dict__.items():
            if isinstance(val, NBContainer):
                yield from val.iter_classes()
            else:
                yield key, val


def nb_print(*args, **kwargs):
    """
    Print output but suppress if "silent".

    Parameters
    ----------
    mssg : Any
        The item/message to show

    """
    if get_opt("verbose") and not get_opt("silent"):
        print(*args, **kwargs)


def nb_data_wait(source: str):
    """
    Print Getting data message.

    Parameters
    ----------
    source : str
        The data source.

    """
    nb_markdown(f"Getting data from {source}...")


def nb_debug(*args):
    """Print debug args."""
    if get_opt("debug"):
        for arg in args:
            print(arg, end="--")
        print()


def nb_markdown(*args, **kwargs):
    """Display Markdown/HTML text."""
    if not get_opt("silent"):
        if _IP_AVAILABLE:
            mp_utils.md(*args, **kwargs)
        else:
            nb_print(*args)


def nb_warn(*args, **kwargs):
    """Display Markdown/HTML warning text."""
    if not get_opt("silent"):
        if _IP_AVAILABLE:
            mp_utils.md_warn(*args, **kwargs)
        else:
            nb_print("WARNING:", *args)


def nb_display(*args, **kwargs):
    """Ipython display function wrapper."""
    if not get_opt("silent"):
        display(*args, **kwargs)


# pylint: disable=invalid-name, unused-argument
def set_text(  # noqa: MC0001
    title: Optional[str] = None,
    hd_level: int = 2,
    text: Optional[str] = None,
    md: bool = False,
    docs: Optional[Dict[str, Any]] = None,
    key: Optional[str] = None,
):
    """
    Decorate function to print title/text before execution.

    Parameters
    ----------
    title : Optional[str], optional
        Title text to print, by default None
    hd_level : int
        Heading level (1-4), by default 2
    text : Optional[str], optional
        Text to print, by default None
    md : bool, optional
        Treat `text` as markdown, by default False
    docs : Dict[str, Any]
        Dictionary of cell documentation indexed by `key`
    key : str
        Item to use from `docs` dictionary.

    Returns
    -------
    Callable[*args, **kwargs]
        Wrapped function

    """

    def text_wrapper(func):
        @functools.wraps(func)
        def print_text(*args, **kwargs):
            # "silent" can be global option or in the func kwargs
            # The latter is only applicable for the NB run() method.
            run_silent = kwargs.get("silent") or get_opt("silent")

            if not run_silent:
                h_level = hd_level
                out_title = title
                out_text = text
                other_items = {}
                if docs and key:
                    out_title = docs.get(key, {}).get("title")
                    out_text = docs.get(key, {}).get("text")
                    h_level = docs.get(key, {}).get("hd_level", 2)
                    other_items = {
                        hdr: str(text)
                        for hdr, text in docs.get(key, {}).items()
                        if hdr not in ("title", "text", "doc", "hd_level", "md")
                    }
                if out_title:
                    h_level = max(min(h_level, 4), 1)
                    display(HTML(f"<h{h_level}>{out_title}</h{h_level}>"))
                if out_text:
                    if md:
                        display(HTML(markdown(text=out_text)))
                    else:
                        display(HTML(out_text.replace("\n", "<br>")))
                if other_items:
                    for sec_title, content in other_items.items():
                        display(HTML(f"<br><b>{sec_title}</b><br>"))
                        display(HTML(content.replace("\n", "<br>")))
            return func(*args, **kwargs)

        return print_text

    return text_wrapper


# pylint: enable=invalid-name, unused-argument


def add_result(result: Any, attr_name: Union[str, List[str]]):
    """
    Decorate func to add return value(s) to `result`.

    Parameters
    ----------
    result : Any
        Object that will have result attributes set.
    attr_name: str or List[str]
        Name of return attribute to set on `result`.

    Returns
    -------
    Callable[*args, **kwargs]
        Wrapped function

    """

    def result_wrapper(func):
        @functools.wraps(func)
        def add_results(*args, **kwargs):
            func_results = func(*args, **kwargs)
            if isinstance(attr_name, str):
                setattr(result, attr_name, func_results)
            elif isinstance(attr_name, list):
                for attr, ret_val in zip(attr_name, func_results):
                    setattr(result, attr, ret_val)
            return func_results

        return add_results

    return result_wrapper


def show_bokeh(plot):
    """Display bokeh plot, resetting output."""
    try:
        bokeh.io.reset_output()
        bokeh.io.output_notebook(hide_banner=True)
        bokeh.io.show(plot)
    except RuntimeError:
        bokeh.io.output_notebook(hide_banner=True)
        bokeh.io.show(plot)


def df_has_data(data) -> bool:
    """Return True if `data` DataFrame has data."""
    return data is not None and not data.empty


class MsticnbError(Exception):
    """Generic exception class for Notebooklets."""


class MsticnbMissingParameterError(MsticnbError):
    """Parameter Error."""

    def __init__(self, *args):
        """
        Exception for missing parameter.

        Parameters
        ----------
        args : str
            First arg is the name or names of the parameters.

        """
        if args:
            self.mssg = f"Required parameter(s) '{args[0]}' not supplied."
            args = (self.mssg, *args[1:])
        super().__init__(*args)


class MsticnbDataProviderError(MsticnbError):
    """DataProvider Error."""


def mp_version():
    """Return currently-loaded msticpy version."""
    return parse(MP_VERSION)


def check_mp_version(required_version: str) -> bool:
    """Return true if the installed version is >= `required_version`."""
    return mp_version().major >= parse(required_version).major


def check_current_result(
    result, attrib: Optional[str] = None, silent: bool = False
) -> bool:
    """
    Check that the result is valid and `attrib` contains data.

    Parameters
    ----------
    result: NotebookletResult
        The result data to check in.
    attrib : str
        Name of the attribute to check, if None this function
    silent : bool
        If True, suppress output.

    Returns
    -------
    bool
        Returns True if valid data is available, else False.

    """
    if not attrib:
        return True
    data_obj = getattr(result, attrib)
    if data_obj is None or isinstance(data_obj, pd.DataFrame) and data_obj.empty:
        if not silent:
            nb_markdown(f"No data is available for {attrib}.")
        return False
    return True



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Notebooklets global options.

Available options are:
[name, type (default value), description]

- `verbose`: bool (True) - Show progress messages.
- `debug`: bool (False) - Turn on debug output.
- `show_sample_results`: bool (True) - Display sample of results as they are produced.
- `silent`: bool (False) - Execute notebooklets with no output.

"""
from typing import Any, Dict

from ._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


_OPTION_DEFN = {
    "verbose": (True, "Show progress messages."),
    "debug": (False, "Turn on debug output."),
    "show_sample_results": (False, "Display sample of results as they are produced."),
    "silent": (False, "Execute notebooklets with no output"),
    "temp_silent": (False, "Execute notebooklets with no output"),
}


_OPT_DICT: Dict[str, Any] = {key: val[0] for key, val in _OPTION_DEFN.items()}


def show():
    """Show help for options."""
    print(
        "\n".join(
            [f"{key} (default={val[0]}): {val[1]}" for key, val in _OPTION_DEFN.items()]
        )
    )


def current():
    """Show current settings."""
    print("\n".join([f"{key}: {val}" for key, val in _OPT_DICT.items()]))


def get_opt(option: str) -> Any:
    """
    Get the named option.

    Parameters
    ----------
    option : str
        Option name.

    Returns
    -------
    Any
        Option value

    Raises
    ------
    KeyError
        An invalid option name was supplied.

    """
    if option in _OPT_DICT:
        if option == "silent":
            if _OPT_DICT.get("temp_silent") is not None:
                return _OPT_DICT.get("temp_silent")
            return _OPT_DICT.get("silent")
        return _OPT_DICT[option]
    raise KeyError(f"Unknown option {option}.")


def set_opt(option: str, value: Any):
    """
    Set the named option.

    Parameters
    ----------
    option : str
        Option name.

    value : Any
        Option value.

    Raises
    ------
    KeyError
        An invalid option name was supplied.
    TypeError
        Option value was not the correct type.

    """
    if option not in _OPT_DICT:
        raise KeyError(f"Unrecognized option {option}.")

    cur_opt = _OPT_DICT[option]
    # allow temp_silent to be None
    if option != "temp_silent" and not isinstance(value, type(cur_opt)):
        try:
            value = type(cur_opt)(value)
        except ValueError as err:
            raise TypeError(
                f"Option is of type {type(cur_opt)}.",
                "{value} cannot be converted to that type.",
            ) from err
    _OPT_DICT[option] = value



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet Result base classes."""
import inspect
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
from bokeh.models import LayoutDOM
from msticpy.common.timespan import TimeSpan

from ._version import VERSION
from .common import show_bokeh
from .data_viewers import DFViewer

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-few-public-methods
class NotebookletResult(DFViewer):
    """Base result class."""

    _TITLE_STYLE = "color:black; background-color:lightgray; padding:5px;"

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional[Any] = None,  # type: ignore
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[Notebooklet], optional
            Originating notebooklet, by default None

        """
        self.description = description or self.__class__.__qualname__
        self.timespan = timespan
        self.notebooklet = notebooklet
        self._attribute_desc: Dict[str, Tuple[str, str]] = {}

        # Populate the `_attribute_desc` dictionary on init.
        self._populate_attr_desc()

    def __str__(self):
        """Return string representation of object."""
        return "\n".join(
            f"{name}: {self._str_repr(val)}"
            for name, val in self.__dict__.items()
            if not name.startswith("_") and val is not None
        )

    @staticmethod
    def _str_repr(obj):
        if isinstance(obj, pd.DataFrame):
            return f"DataFrame: {len(obj)} rows"
        if isinstance(obj, LayoutDOM):
            return "Bokeh plot"
        return str(obj)

    # pylint: disable=unsubscriptable-object, no-member
    def _repr_html_(self):
        """Display HTML represention for notebook."""
        attrib_lines = []
        for name, val in self.__dict__.items():
            if name.startswith("_") or val is None:
                continue
            attr_desc = ""
            attr_type, attr_text = self._attribute_desc.get(
                name, (None, None)
            )  # type: ignore
            if attr_text:
                attr_desc += f"{attr_text}"
            if attr_type:
                attr_desc += f"&nbsp;Type: [{attr_type}]"
            attrib_lines.extend(
                [
                    f"<h3 style='{self._TITLE_STYLE}'>{name}</h3>",
                    f"{attr_desc}<br>{self._html_repr(val)}<hr>",
                ]
            )
        return "".join(attrib_lines)

    # pylint: enable=unsubscriptable-object, no-member

    # pylint: disable=protected-access
    @staticmethod
    def _html_repr(obj):
        if isinstance(obj, pd.DataFrame):
            suffix = f"<br>(showing top 5 of {len(obj)} rows)" if len(obj) > 5 else ""
            return obj.head(5)._repr_html_() + suffix
        if isinstance(obj, LayoutDOM):
            show_bokeh(obj)
        if hasattr(obj, "_repr_html_"):
            return obj._repr_html_()
        return str(obj).replace("\n", "<br>").replace(" ", "&nbsp;")

    # pylint: enable=protected-access

    def __getattr__(self, name):
        """Proxy attributes of the notebooklet member."""
        if self.notebooklet:
            return getattr(self.notebooklet, name)
        raise AttributeError(f"{self.__class__} has no attribute '{name}'")

    def _populate_attr_desc(self):
        indent = " " * 4
        in_attribs = False
        attr_name = None
        attr_type = None
        attr_dict = {}
        attr_lines = []
        doc_str = inspect.cleandoc(self.__doc__)
        for line in doc_str.split("\n"):
            if line.strip() == "Attributes":
                in_attribs = True
                continue
            if (
                line.strip() == "-" * len("Attributes")
                or not in_attribs
                or not line.strip()
            ):
                continue
            if not line.startswith(indent):
                # if existing attribute, add to dict
                if attr_name:
                    attr_dict[attr_name] = attr_type, " ".join(attr_lines)
                if ":" in line:
                    attr_name, attr_type = [item.strip() for item in line.split(":")]
                else:
                    attr_name = line.strip()
                    attr_type = "object"
                attr_lines = []
            else:
                attr_lines.append(line.strip())
        attr_dict[attr_name] = attr_type, " ".join(attr_lines)
        if "timespan" not in attr_dict:
            attr_dict["timespan"] = (
                "TimeSpan",
                "Time span for the queried results data.",
            )
        if "notebooklet" not in attr_dict:
            attr_dict["notebooklet"] = (
                "Notebooklet",
                "The notebooklet instance that created this result.",
            )
        # pylint: disable=no-member
        self._attribute_desc.update(attr_dict)  # type: ignore
        # pylint: enable=no-member

    @property
    def properties(self):
        """Return names of all properties."""
        return [
            name
            for name, val in self.__dict__.items()
            if val is not None and not name.startswith("_")
        ]

    def prop_doc(self, name) -> Tuple[str, str]:
        """Get the property documentation for the property."""
        # pylint: disable=unsupported-membership-test, unsubscriptable-object
        if name in self._attribute_desc:
            return self._attribute_desc[name]
        # pylint: enable=unsupported-membership-test, unsubscriptable-object
        raise KeyError(f"Unknown property {name}.")

    def data_properties(self, empty: bool = False) -> List[str]:
        """Return list of attributes with populated data."""
        return [
            attr
            for attr, val in vars(self).items()
            if isinstance(val, pd.DataFrame) and (empty or not val.empty)
        ]

    def vis_properties(self) -> List[str]:
        """Return list of properties with visualizations."""
        return [attr for attr, val in vars(self).items() if isinstance(val, LayoutDOM)]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""NB metadata test class."""
from msticnb import browse, init
from msticnb.notebooklet import Notebooklet


def test_nb_browse():
    """Test Notebooklet browser."""
    init()
    browser = browse()

    nb_list = browser.nb_select.options
    assert len(nb_list) > 6
    cls_name, cls_obj = nb_list[0]
    browser.nb_select.value = cls_obj

    selected_class = browser.selected
    assert cls_obj == selected_class
    assert browser.nb_doc.value is not None
    assert cls_obj.__name__ == cls_name
    assert cls_obj.__name__ in browser.nb_doc.value
    assert browser.nb_run_doc.value is not None
    assert browser.nb_code.value
    nb_doc = browser.nb_doc.value is not None
    nb_run_doc = browser.nb_run_doc.value
    nb_code = browser.nb_code.value

    # Select a new option
    cls_name, cls_obj = nb_list[1]
    browser.nb_select.value = cls_obj

    assert nb_list
    selected_class = browser.selected
    assert issubclass(selected_class, Notebooklet)
    assert browser.nb_doc.value is not None
    assert cls_obj.__name__ in browser.nb_doc.value
    assert browser.nb_run_doc.value is not None
    assert browser.nb_code.value is not None

    assert browser.nb_doc.value != nb_doc
    assert browser.nb_run_doc.value != nb_run_doc
    assert browser.nb_code.value != nb_code



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Logon sessions rarity analysis."""
from typing import Any, Dict, Iterable, Optional

import pandas as pd
from msticpy.analysis.eventcluster import char_ord_score, dbcluster_events, delim_count
from msticpy.common.timespan import TimeSpan

# pylint: disable=unused-import
from msticpy.init import mp_pandas_accessors  # noqa: F401

try:
    from msticpy import nbwidgets

    if not hasattr(nbwidgets, "SelectItem"):
        raise ImportError("Invalid nbwidgets")
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools import nbwidgets

from .... import nb_metadata
from ...._version import VERSION
from ....common import (
    MsticnbMissingParameterError,
    nb_display,
    nb_markdown,
    nb_print,
    set_text,
)
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Ian Hellen"


# Read module metadata from YAML
_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = nb_metadata.read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods
# Rename this class
class LogonSessionsRarityResult(NotebookletResult):
    """
    Logon Sessions rarity.

    Attributes
    ----------
    process_clusters : pd.DataFrame
        Process clusters based on account, process, commandline and
        showing the an example process from each cluster
    processes_with_cluster : pd.DataFrame
        Merged data with rarity value assigned to each process event.
    session_rarity: pd.DataFrame
        List of sessions with averaged process rarity.

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[, optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.description: str = "Windows Host Security Events"

        # Add attributes as needed here.
        # Make sure they are documented in the Attributes section
        # above.
        self.process_clusters: Optional[pd.DataFrame] = None
        self.processes_with_cluster: Optional[pd.DataFrame] = None
        self.session_rarity: Optional[pd.DataFrame] = None


# pylint: enable=too-few-public-methods


# Rename this class
class LogonSessionsRarity(Notebooklet):
    """
    Calculates the relative rarity of logon sessions.

    It clusters the data based on process, command line and account.
    Then calculates the rarity of the process pattern.
    Then returns a result containing a summary of the logon sessions by rarity.

    To see the methods available for the class and result class, run
    cls.list_methods()

    """

    # assign metadata from YAML to class variable
    metadata = _CLS_METADATA
    __doc__ = nb_metadata.update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    def __init__(self, **kwargs):
        """Initialize instance of LogonSessionRarity."""
        super().__init__(**kwargs)
        self.column_map = {}
        self._event_browser = None

    # @set_text decorator will display the title and text every time
    # this method is run.
    # The key value refers to an entry in the `output` section of
    # the notebooklet yaml file.
    @set_text(docs=_CELL_DOCS, key="run")
    def run(
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> LogonSessionsRarityResult:
        """
        Calculate Logon sessions ordered by process rarity summary.

        Parameters
        ----------
        value : str
            Not used
        data : Optional[pd.DataFrame], optional
            Process event data.
        timespan : TimeSpan
            Not used
        options : Optional[Iterable[str]], optional
            List of options to use, by default None.
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Returns
        -------
        LogonSessionsRarityResult
            LogonSessionsRarityResult.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        # This line use logic in the superclass to populate options
        # (including default options) into this class.
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if data is None:
            raise MsticnbMissingParameterError("data")

        # Create a result class
        result = LogonSessionsRarityResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )

        self.column_map = _get_column_map(data)
        feat_data, cols = _add_session_features(data=data, column_map=self.column_map)
        result.process_clusters, labeled_events = _cluster_sessions(
            data=feat_data, columns=list(cols.keys())
        )

        (
            result.processes_with_cluster,
            result.session_rarity,
        ) = _merge_cluster_with_procs(
            data=labeled_events,
            clustered_data=result.process_clusters,
            column_map=self.column_map,
        )
        # save the result
        self._last_result = result

        if not self.silent:
            self.list_sessions_by_rarity()
            self.plot_sessions_by_rarity()

        self._event_browser = _create_session_browser(
            summ_data=result.session_rarity,
            data=result.processes_with_cluster,
            column_map=self.column_map,
        )

        nb_markdown("<h3>View the returned results object for more details.</h3>")
        nb_markdown(
            f"Additional methods for this class:<br>{'<br>'.join(self.list_methods())}"
        )
        return self._last_result

    def list_sessions_by_rarity(self):
        """Display sessions by process rarity."""
        if self.check_valid_result_data("session_rarity"):
            nb_display(
                self._last_result.session_rarity.sort_values(
                    "MeanRarity", ascending=False
                ).style.bar(subset=["MeanRarity"], color="#d65f5f")
            )

    def plot_sessions_by_rarity(self):
        """Display timeline plot of processes by rarity."""
        if self.check_valid_result_data("processes_with_cluster"):
            data = self._last_result.processes_with_cluster
            acct_col = self.column_map.get(COL_ACCT)
            data.mp_plot.timeline_values(
                title="Processes with relative rarity score grouped by Account",
                y="Rarity",
                group_by=acct_col,
                height=600,
                kind=["vbar", "circle"],
                source_columns=[self.column_map[COL_PROC], self.column_map[COL_CMD]],
            )

    def process_tree(
        self, account: Optional[str] = None, session: Optional[str] = None
    ):
        """
        View a process tree of current session.

        Parameters
        ----------
        account : Optional[str], optional
            The account name to view, by default None
        session : Optional[str], optional
            The logon session to view, by default None

        """
        if self.check_valid_result_data("processes_with_cluster"):
            if (not account and not session) or account == "all":
                self._last_result.processes_with_cluster.mp_plot.process_tree(
                    legend_col="Rarity"
                )
                return
            if account:
                acct_col = self.column_map.get(COL_ACCT)
                data = self._last_result.processes_with_cluster
                data = data[data[acct_col] == account]
                proc_tree_data = data.mp.build_process_tree()
                proc_tree_data["Rarity"] = pd.to_numeric(
                    proc_tree_data["Rarity"], errors="coerce"
                ).fillna(0)
                proc_tree_data.mp_plot.process_tree(legend_col="Rarity")
                return
            session = session or self._event_browser.value
            sess_col = self.column_map.get(COL_SESS)
            data = self._last_result.processes_with_cluster
            data = data[data[sess_col] == session]
            proc_tree_data = data.mp.build_process_tree()
            proc_tree_data["Rarity"] = pd.to_numeric(
                proc_tree_data["Rarity"], errors="coerce"
            ).fillna(0)
            proc_tree_data.mp_plot.process_tree(legend_col="Rarity")

    def browse_events(self):
        """Browse the events by logon session."""
        if self.check_valid_result_data("processes_with_cluster"):
            return self._event_browser
        return None


# %
# Get the column mapping for the data
COL_ACCT = "acct"
COL_TS = "timestamp"
COL_PROC = "process_name"
COL_CMD = "command"
COL_SESS = "sess"
COL_PID = "pid"


def _find_column(data, column_opts, default=None):
    for col in column_opts:
        if col in data.columns:
            return col
    return default


def _get_column_map(data):
    return {
        COL_ACCT: _find_column(data, ["Account", "SubjectLogonName", "acct", "uid"]),
        COL_TS: _find_column(data, ["TimeGenerated", "EventStartTime", "TimeStamp"]),
        COL_CMD: _find_column(data, ["CommandLine", "cmd"]),
        COL_PROC: _find_column(data, ["NewProcessName", "exe"]),
        COL_SESS: _find_column(data, ["SubjectLogonId", "ses"]),
        COL_PID: _find_column(data, ["NewProcessId", "pip"]),
    }


CMD_LINE_TOKS = "CommandlineTokensFull"
PATH_SCORE = "PathScore"
ACCT_NUM = "AccountNum"
SYS_SESS = "IsSystemSession"
CLUSTER_COLUMNS = [CMD_LINE_TOKS, PATH_SCORE, ACCT_NUM, SYS_SESS]


# %%
def _add_session_features(data, column_map: Dict[str, str]):
    """Create clustering features."""
    nb_markdown(f"Input data: {len(data)} events")
    nb_print("Extracting features...", end="")
    data = data.copy()

    cluster_cols = {}
    proc_name = column_map.get(COL_PROC)
    if proc_name:
        cluster_cols[PATH_SCORE] = proc_name
        data[PATH_SCORE] = data.apply(lambda x: char_ord_score(x[proc_name]), axis=1)
        nb_print(".", end="")
    cmd_line = column_map.get("command", "CommandLine")
    if cmd_line:
        cluster_cols[CMD_LINE_TOKS] = cmd_line
        data[CMD_LINE_TOKS] = data.apply(lambda x: delim_count(x[cmd_line]), axis=1)
        nb_print(".", end="")
    acct = column_map.get("account", "Account")
    if acct:
        cluster_cols[ACCT_NUM] = acct
        data[ACCT_NUM] = data.apply(lambda x: char_ord_score(x[acct]), axis=1)
        nb_print(".", end="")
    sess = column_map.get("logon_id", "SubjectLogonId")
    if sess:
        cluster_cols[SYS_SESS] = sess
        data[SYS_SESS] = data[sess].isin(["0x3e7", "-1"])
    nb_markdown("done.")
    return data, cluster_cols


def _cluster_sessions(data, columns=None):
    """Cluster data using DBSCAN."""
    # you might need to play around with the max_cluster_distance parameter.
    # decreasing this gives more clusters.
    columns = columns or CLUSTER_COLUMNS
    nb_markdown("Clustering...")
    (clus_events, db_cluster, _) = dbcluster_events(
        data=data,
        cluster_columns=columns,
        max_cluster_distance=0.0001,
    )
    labeled_events = data
    labeled_events["ClusterId"] = db_cluster.labels_
    nb_markdown("done")
    nb_markdown(f"Number of input events: {len(data)}")
    nb_markdown(
        f"Number of clusters: {len(clus_events[clus_events['ClusterId'] != -1])}"
    )
    nb_markdown(
        "Number of unique (unclustered) events: "
        f"{len(clus_events[clus_events['ClusterId'] == -1])}"
    )
    return clus_events, labeled_events


def _merge_cluster_with_procs(data, clustered_data, column_map):
    """Merge clustered data with original."""
    nb_markdown("Merging with source data and computing rarity...")

    # Join the clustered results back to the original process frame
    noise_points = data[data["ClusterId"] == -1].assign(ClusterId=-1, ClusterSize=1)
    clusters = data[data["ClusterId"] != -1].merge(
        clustered_data[["ClusterId", "ClusterSize"]],
        on="ClusterId",
    )
    procs_with_cluster = pd.concat([clusters, noise_points])
    # Compute Process pattern Rarity = inverse of cluster size
    procs_with_cluster["Rarity"] = 1 / procs_with_cluster["ClusterSize"]

    # count the number of processes for each logon ID
    sess = column_map.get(COL_SESS, "SubjectLogonId")
    acct = column_map.get(COL_ACCT, "Account")
    timestamp = column_map.get(COL_TS, "TimeGenerated")
    session_rarity = (
        procs_with_cluster.groupby([acct, sess])
        .agg(
            MeanRarity=pd.NamedAgg("Rarity", "mean"),
            MaxRarity=pd.NamedAgg("Rarity", "max"),
            ProcessCount=pd.NamedAgg(timestamp, "count"),
        )
        .reset_index()
    )

    nb_markdown("done")
    # Display the results
    nb_markdown("<br><hr>Sessions ordered by process rarity", "large, bold")
    nb_markdown("Higher score indicates higher number of unusual processes")

    return procs_with_cluster, session_rarity


def _create_session_browser(summ_data, data, column_map):
    browse_cols = [
        column_map[COL_ACCT],
        column_map[COL_TS],
        column_map[COL_SESS],
        column_map[COL_PID],
        column_map[COL_PROC],
        column_map[COL_CMD],
    ]
    if "ParentProcessName" in data:
        browse_cols.append("ParentProcessName")
    browse_cols.append("Rarity")

    item_dict = {
        f"{item[1]} - {item[0]}, mean rarity: {item[2]}": item[0]
        for item in summ_data[
            [column_map[COL_SESS], column_map[COL_ACCT], "MeanRarity"]
        ].values
    }

    def show_events(logon_id):
        return (
            data[browse_cols]
            .query(f"{column_map[COL_SESS]} == '{logon_id}'")
            .sort_values(column_map[COL_TS])
        )

    return nbwidgets.SelectItem(item_dict=item_dict, action=show_events)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet for URL Summary."""
from collections import Counter
from os.path import exists
from typing import Any, Dict, Iterable, List, Optional

import dns.resolver
import numpy as np
import pandas as pd
import tldextract
from IPython.display import Image, display

# pylint: disable=ungrouped-imports
try:
    from msticpy import nbwidgets
    from msticpy.context.domain_utils import DomainValidator, screenshot
    from msticpy.context.ip_utils import ip_whois as whois
    from msticpy.vis.timeline import display_timeline, display_timeline_values
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from whois import whois  # type: ignore
    from msticpy.sectools.domain_utils import DomainValidator, screenshot
    from msticpy.nbtools import nbwidgets
    from msticpy.nbtools.nbdisplay import display_timeline, display_timeline_values

from bokeh.models import LayoutDOM
from msticpy.common.timespan import TimeSpan
from msticpy.common.utility import md

from ...._version import VERSION
from ....common import (
    MsticnbDataProviderError,
    MsticnbMissingParameterError,
    nb_markdown,
    set_text,
)
from ....nb_metadata import read_mod_metadata, update_class_doc
from ....nblib.azsent.alert import browse_alerts
from ....nblib.ti import get_ti_results
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Pete Bryan"


_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods, too-many-instance-attributes
class URLSummaryResult(NotebookletResult):
    """URL Summary Results."""

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """Create new Notebooklet result instance."""
        super().__init__(description, timespan, notebooklet)
        self.summary: Optional[pd.DataFrame] = None
        self.domain_record: Optional[pd.DataFrame] = None
        self.cert_details: Optional[pd.DataFrame] = None
        self.ip_record: Optional[pd.DataFrame] = None
        self.related_alerts: Optional[pd.DataFrame] = None
        self.bookmarks: Optional[pd.DataFrame] = None
        self.dns_results: Optional[pd.DataFrame] = None
        self.hosts: Optional[List] = None
        self.flows: Optional[pd.DataFrame] = None
        self.flow_graph: Optional[LayoutDOM] = None
        self.ti_results: Optional[pd.DataFrame] = None


# pylint: disable=too-few-public-methods
class URLSummary(Notebooklet):
    """
    URLSummary Notebooklet class.

    Queries and displays information about a URL including:

    - Overview of URL and Domain
    - Related alerts
    - TI results

    """

    metadata = _CLS_METADATA
    __doc__ = update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    # pylint: disable=too-many-branches, too-many-locals, too-many-statements
    @set_text(docs=_CELL_DOCS, key="run")  # noqa: MC0001
    def run(  # noqa:MC0001, C901
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> URLSummaryResult:
        """
        Return host summary data.

        Parameters
        ----------
        value : str
            URL to investigate
        data : Optional[pd.DataFrame], optional
            Not used, by default None
        timespan : TimeSpan
            Timespan over which operations such as queries will be
            performed, by default None.
            This can be a TimeStamp object or another object that
            has valid `start`, `end`, or `period` attributes.
        options : Optional[Iterable[str]], optional
            List of options to use, by default None
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Other Parameters
        ----------------
        start : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.
        end : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.

        Returns
        -------
        URLSummaryResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not value:
            raise MsticnbMissingParameterError("value")
        if not timespan:
            raise MsticnbMissingParameterError("timespan.")

        self.timespan = timespan

        # pylint: disable=attribute-defined-outside-init
        result = URLSummaryResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )
        if not self._last_result:
            self._last_result = result

        self.url = value.strip().lower()

        extracted_result = tldextract.extract(self.url)
        domain = extracted_result.registered_domain

        domain_validator = DomainValidator()
        validated = domain_validator.validate_tld(domain)

        result.summary = pd.DataFrame(
            {"URL": [self.url], "Domain": [domain], "Validated TLD": [validated]}
        )

        if "ti" in self.options:
            if "tilookup" in self.data_providers.providers:
                ti_prov = self.data_providers.providers["tilookup"]
            else:
                raise MsticnbDataProviderError("No TI providers available")
            ti_results, ti_results_merged = get_ti_results(
                ti_prov, result.summary, "URL"
            )
            if isinstance(ti_results, pd.DataFrame) and not ti_results.empty:
                result.ti_results = ti_results_merged

        if "whois" in self.options:
            result.domain_record = _domain_whois_record(
                domain, self.data_providers.providers["tilookup"]
            )

        if "cert" in self.options:
            result.cert_details = _get_tls_cert_details(self.url, domain_validator)

        if "ip_record" in self.options:
            result.ip_record = None
            result.ip_record = _get_ip_record(
                domain, domain_validator, self.data_providers.providers["tilookup"]
            )

        if "screenshot" in self.options:
            image_data = screenshot(self.url)
            with open("screenshot.png", "wb") as screenshot_file:
                screenshot_file.write(image_data.content)

        if "alerts" in self.options:
            alerts = self.query_provider.SecurityAlert.list_alerts(timespan)
            result.related_alerts = alerts[
                alerts["Entities"].str.contains(self.url, case=False)
            ]

        if "bookmarks" in self.options:
            result.bookmarks = (
                self.query_provider.AzureSentinel.list_bookmarks_for_entity(
                    url=self.url, start=timespan.start, end=timespan.end
                )
            )
        available_queries = set(self.query_provider.list_queries())
        if (
            "dns" in self.options
            and "AzureNetwork.dns_lookups_for_domain" in available_queries
        ):
            result.dns_results = (
                self.query_provider.AzureNetwork.dns_lookups_for_domain(
                    domain=domain, start=timespan.start, end=timespan.end
                )
            )

        if "hosts" in self.options:
            syslog_hosts = []
            mde_hosts = []
            windows_hosts = []
            if "LinuxSyslog.all_syslog" in available_queries:
                syslog_hosts = self.query_provider.LinuxSyslog.all_syslog(
                    add_query_items=f"| where SyslogMessage has '{self.url}'",
                    start=timespan.start,
                    end=timespan.end,
                )["Computer"].unique()
            if "MDE.host_connections" in available_queries:
                mde_hosts_df = self.query_provider.MDE.host_connections(
                    time_column="TimeGenerated",
                    host_name="",
                    add_query_items=f"| where RemoteUrl has '{self.url}'",
                    start=timespan.start,
                    end=timespan.end,
                )
                if "DeviceName" in mde_hosts_df.columns:
                    mde_hosts = mde_hosts_df["DeviceName"].unique()
                elif "Computer" in mde_hosts_df.columns:
                    mde_hosts = mde_hosts_df["Computer"].unique()
            if "WindowsSecurity.list_events" in available_queries:
                windows_hosts = self.query_provider.WindowsSecurity.list_events(
                    add_query_items=f"| where CommandLine has '{self.url}'",
                    start=timespan.start,
                    end=timespan.end,
                )["Computer"].unique()
            all_hosts = list(syslog_hosts) + list(mde_hosts) + list(windows_hosts)
            result.hosts = all_hosts

        if "flows" in self.options:
            if "Network.network_connections_to_url" in available_queries:
                result.flows = self.query_provider.Network.network_connections_to_url(
                    start=timespan.start, end=timespan.end, url="com"
                )
                flow_graph_data = self.query_provider.Network.network_connections_to_url(
                    start=timespan.start,
                    end=timespan.end,
                    url="com",
                    add_query_items=(
                        "| summarize sum(SentBytes) by RequestURL, bin(TimeGenerated, 10m)"
                    ),
                )
                result.flow_graph = display_timeline_values(
                    flow_graph_data,
                    value_col="sum_SentBytes",
                    title=f"Network traffic volume to {self.url}",
                )

        self._last_result = result

        if not self.silent:
            self._display_results()

        return self._last_result

    @set_text(docs=_CELL_DOCS, key="display_summary")
    def _display_summary(self):
        """Display URL summary."""
        if self.check_valid_result_data("summary", silent=True):
            display(self._last_result.summary)

    @set_text(docs=_CELL_DOCS, key="show_ti_details")
    def _display_ti_data(self):
        """Display TI results."""
        if self.check_valid_result_data("ti_results", silent=True):
            display(self._last_result.ti_results)
        else:
            nb_markdown(f"No TI results found for {self.url}")

    @set_text(docs=_CELL_DOCS, key="show_domain_record")
    def _display_domain_record(self):
        """Display Domain Record."""
        if self.check_valid_result_data("domain_record", silent=True):
            display(
                self._last_result.domain_record.T.style.map(  # type: ignore
                    color_domain_record_cells,
                    subset=pd.IndexSlice[["Page Rank", "Domain Name Entropy"], 0],
                )
            )

    @set_text(docs=_CELL_DOCS, key="show_TLS_cert")
    def _display_cert_details(self):
        """Display TLS Certificate details."""
        if self.check_valid_result_data("cert_details", silent=True):
            display(self._last_result.cert_details)
        else:
            nb_markdown(f"No TLS certificate found for {self.url}.")

    @set_text(docs=_CELL_DOCS, key="show_IP_record")
    def _display_ip_record(self):
        """Display IP record."""
        if self.check_valid_result_data("ip_record", silent=True):
            display(self._last_result.ip_record.T)
        else:
            nb_markdown(f"No current IP found for {self.url}.")

    @set_text(docs=_CELL_DOCS, key="show_screenshot")
    def _display_screenshot(self):
        """Display ULR screenshot."""
        if exists("screenshot.png"):
            nb_markdown(f"Screenshot of {self.url}")
            display(Image(filename="screenshot.png"))

    @set_text(docs=_CELL_DOCS, key="show_related_alerts")
    def _display_related_alerts(self):
        """Display related alerts in table."""
        if self.check_valid_result_data("related_alerts", silent=True):
            display(self._last_result.related_alerts)
        else:
            nb_markdown(f"No Alerts related to {self.url}")

    @set_text(docs=_CELL_DOCS, key="show_bookmarks")
    def _display_bookmarks(self):
        """Display bookmarks related to URL."""
        if self.check_valid_result_data("bookmarks", silent=True):
            display(self._last_result.bookmarks)
        else:
            nb_markdown(f"No Bookmarks related to {self.url}")

    @set_text(docs=_CELL_DOCS, key="show_dns_results")
    def _display_dns_results(self):
        """Display DNS resolutions for URL."""
        if self.check_valid_result_data("dns_results", silent=True):
            nb_markdown(f"DNS events related to {self.url}", "bold")
            display(self._last_result.dns_results)
        else:
            nb_markdown(f"No DNS resolutions found for {self.url}")

    @set_text(docs=_CELL_DOCS, key="show_hosts")
    def _display_hosts(self):
        """Display list of hosts connecting to URL."""
        if (
            self.check_valid_result_data("hosts", silent=True)
            and self._last_result.hosts
        ):
            nb_markdown(f"Hosts connecting to {self.url}", "bold")
            display(self._last_result.hosts)
        else:
            nb_markdown(f"No hosts found connecting to {self.url}")

    @set_text(docs=_CELL_DOCS, key="show_flows")
    def _display_flows(self):
        """Display network flow data for URL."""
        if self.check_valid_result_data("flow_graph", silent=True):
            display(self._last_result.flow_graph)
            nb_markdown(f"Network connections to {self.url}", "bold")
            display(self._last_result.flows)
        else:
            nb_markdown(f"No flow data found for {self.url}")

    @set_text(docs=_CELL_DOCS, key="browse_alerts")
    def browse_alerts(self) -> nbwidgets.SelectAlert:
        """Return alert browser/viewer."""
        if self.check_valid_result_data("related_alerts"):
            return browse_alerts(self._last_result)
        return None

    def display_alert_timeline(self):
        """Display the alert timeline."""
        if self.check_valid_result_data("related_alerts"):
            if len(self._last_result.related_alerts) > 1:
                return _show_alert_timeline(self._last_result.related_alerts)
            md("Cannot plot timeline with 0 or 1 event.")
        return None

    def _display_results(self):
        """Display all the notebooklet results."""
        self._display_summary()
        self._display_domain_record()
        self._display_ip_record()
        self._display_cert_details()
        self._display_ti_data()
        self._display_screenshot()
        self._display_related_alerts()
        self._display_bookmarks()
        self._display_dns_results()
        self._display_hosts()
        self._display_flows()


def entropy(data):
    """Calculate Entropy of a String."""
    strings, lens = Counter(data), np.float(len(data))
    return -sum(count / lens * np.log2(count / lens) for count in strings.values())


def color_domain_record_cells(val):
    """Color Code Rows With Odd Entropy."""
    if isinstance(val, int):
        color = "yellow" if val < 3 else None
    elif isinstance(val, float):
        color = "yellow" if val > 4.30891 or val < 2.72120 else None
    else:
        color = None
    return f"background-color: {color}"


@set_text(docs=_CELL_DOCS, key="display_alert_timeline")
def _show_alert_timeline(related_alerts):
    """Display An Alert Timeline."""
    if len(related_alerts) > 1:
        return display_timeline(
            data=related_alerts,
            title="Related Alerts",
            source_columns=["AlertName", "TimeGenerated"],
            height=200,
        )
    if len(related_alerts) == 1:
        nb_markdown("A single alert cannot be plotted on a timeline.")
    else:
        nb_markdown("No alerts available to be plotted.")
    return None


def _domain_whois_record(domain, ti_prov):
    """Build a Domain Whois Record."""
    dom_record = pd.DataFrame()
    whois_result = whois(domain)
    if hasattr(whois_result, "domain_name") and whois_result.domain_name is not None:
        # Create domain record from whois data
        dom_record = pd.DataFrame(
            {
                "Domain": [domain],
                "Name": [whois_result.get("name", None)],
                "Org": [whois_result.get("org", None)],
                "DNSSec": [whois_result.get("dnssec", None)],
                "City": [whois_result.get("city", None)],
                "State": [whois_result.get("state", None)],
                "Country": [whois_result.get("country", None)],
                "Registrar": [whois_result.get("registrar", None)],
                "Status": [whois_result.get("status", None)],
                "Created": [whois_result.get("creation_date", None)],
                "Expiration": [whois_result.get("expiration_date", None)],
                "Last Updated": [whois_result.get("updated_date", None)],
                "Name Servers": [whois_result.get("name_servers", None)],
            }
        )
        ns_domains = []

        # Identity domains popularity with Open Page Rank
        if "OPR" in ti_prov.loaded_providers:
            page_rank = ti_prov.result_to_df(
                ti_prov.lookup_ioc(domain, providers=["OPR"])
            )
            if page_rank["RawResult"][0]:
                page_rank_score = page_rank["RawResult"][0]["response"][0][
                    "page_rank_integer"
                ]
            else:
                page_rank_score = 0
            dom_record["Page Rank"] = [page_rank_score]
        else:
            nb_markdown("OPR TI provider needed to calculate Page Rank score.")
            dom_record["Page Rank"] = ["Not known - OPR provider needed"]

        # Get a list of subdomains for the domain
        if "VirusTotal" in ti_prov.loaded_providers:
            url_ti = ti_prov.result_to_df(
                ti_prov.lookup_ioc(domain, providers=["VirusTotal"])
            )
            try:
                sub_doms = url_ti["RawResult"][0]["subdomains"]
            except (TypeError, KeyError):
                sub_doms = "None found"
            dom_record["Sub Domains"] = [sub_doms]
        else:
            nb_markdown("VT TI provider needed to get sub-domains.")
            dom_record["Page Rank"] = ["Not known - OPR provider needed"]

        # Work out domain entropy to identity possible DGA
        dom_ent = entropy(domain)
        dom_record["Domain Name Entropy"] = [dom_ent]

        # Remove duplicate Name Server records
        for server in whois_result["name_servers"]:
            # pylint: disable=unpacking-non-sequence
            _, ns_domain, ns_tld = tldextract.extract(server)
            ns_dom = ns_domain.lower() + "." + ns_tld.lower()
            if domain not in ns_domains:
                ns_domains.append(ns_dom)
    return dom_record


def _get_tls_cert_details(url, domain_validator):
    """Get details of a TLS certificate used by a domain."""
    result, x509 = domain_validator.in_abuse_list(url)
    cert_df = pd.DataFrame()
    if x509 is not None:
        cert_df = pd.DataFrame(
            {
                "SN": [x509.serial_number],
                "Subject": [[(i.value) for i in x509.subject]],
                "Issuer": [[(i.value) for i in x509.issuer]],
                "Expired": [x509.not_valid_after],
                "InAbuseList": result,
            }
        )
    return cert_df


def _get_ip_record(domain, domain_validator, ti_prov):
    """Get IP addresses associated with a domain."""
    ip_record = None
    if domain_validator.is_resolvable(domain) is True:
        try:
            answer = dns.resolver.resolve(domain)
        except dns.resolver.NXDOMAIN:
            md("Could not resolve IP addresses from domain.")
        resolved_domain_ip = answer[0].to_text()
        ip_whois_result = whois(resolved_domain_ip)
        ip_record = pd.DataFrame(
            {
                "IP Address": [resolved_domain_ip],
                "Domain": [ip_whois_result.properties.get("domain_name", None)],
                "Registrar": [ip_whois_result.properties.get("asn_description", None)],
                "Country": [ip_whois_result.properties.get("country", None)],
                "Creation Date": [
                    ip_whois_result.properties.get("creation_date", None)
                ],
            }
        )
    if isinstance(ip_record, pd.DataFrame) and not ip_record.empty:
        ip_record = _process_tor_ip_record(ip_record, ti_prov)
        ip_record = _process_previous_resolutions(ip_record, ti_prov)
    return ip_record


def _process_tor_ip_record(ip_record, ti_prov):
    """See if IP record contains Tor IP."""
    tor = None
    if "Tor" in ti_prov.loaded_providers:
        print(ti_prov.loaded_providers)
        tor = ti_prov.result_to_df(
            ti_prov.lookup_ioc(ip_record["IP Address"][0], providers=["Tor"])
        )
    if tor is None or tor["Details"][0] == "Not found.":
        ip_record["Tor Node?"] = "No"
    else:
        ip_record["Tor Node?"] = "Yes"
    return ip_record


def _process_previous_resolutions(ip_record, ti_prov):
    """Get previous resolutions for IP in ip_record."""
    if "VirusTotal" in ti_prov.loaded_providers:
        ip_ti_results = ti_prov.result_to_df(
            ti_prov.lookup_ioc(ip_record["IP Address"][0], providers=["VirusTotal"])
        )
        prev_domains = None
        try:
            if ip_ti_results is not None and not ip_ti_results.empty:
                last_10 = ip_ti_results.iloc[0]["RawResult"].get("resolutions", [])
                if isinstance(last_10, list):
                    prev_domains = [record["hostname"] for record in last_10[:10]]
        except (TypeError, KeyError, AttributeError):
            prev_domains = None
    else:
        prev_domains = (
            "Unknown - VirusTotal provider required for previous resolution details."
        )
    ip_record["Last 10 resolutions"] = [prev_domains]
    return ip_record



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet for Account Summary."""
from enum import Flag, auto
from typing import Any, Callable, Dict, Iterable, Optional, Union

import pandas as pd
from bokeh.models import LayoutDOM
from IPython.display import HTML
from msticpy.common.timespan import TimeSpan
from msticpy.common.utility import enum_parse
from msticpy.datamodel import entities

try:
    from msticpy import nbwidgets
    from msticpy.vis.timeline import display_timeline
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools import nbwidgets
    from msticpy.nbtools.nbdisplay import display_timeline

from .... import nb_metadata
from ...._version import VERSION
from ....common import (
    MsticnbMissingParameterError,
    df_has_data,
    nb_data_wait,
    nb_display,
    nb_markdown,
    set_text,
)
from ....nblib.azsent.alert import browse_alerts
from ....nblib.iptools import get_geoip_whois, map_ips
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Ian Hellen"

# pylint: disable=too-many-lines
# Read module metadata from YAML
_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = nb_metadata.read_mod_metadata(__file__, __name__)


# pylint: disable=invalid-name
class AccountType(Flag):
    """Account types."""

    AzureActiveDirectory = auto()
    AzureActivity = auto()
    Office365 = auto()
    Windows = auto()
    Linux = auto()
    # pylint:disable=unsupported-binary-operation
    Azure = AzureActiveDirectory | AzureActivity | Office365
    All = Azure | Windows | Linux  # pylint:disable=unsupported-binary-operation

    def in_list(self, acct_types: Iterable[Union["AccountType", str]]):
        """Is the current value in the `acct_types` list."""
        for acct_type in acct_types:
            if isinstance(acct_type, str):
                acct_type = AccountType[acct_type]
            if self & acct_type:
                return True
        return False

    @classmethod
    def parse(cls, name: str):
        """Try to parse string to valid account type."""
        return enum_parse(cls, name)


# pylint: enable=invalid-name

# pylint: disable=no-member


# pylint: disable=too-few-public-methods, too-many-instance-attributes
class AccountSummaryResult(NotebookletResult):
    """
    Account Summary Result.

    Attributes
    ----------
    account_activity : pd.DataFrame
        DataFrame of most recent activity.
    account_selector : msticpy.nbtools.nbwidgets.SelectString
        Selection widget for accounts.
    related_alerts : pd.DataFrame
        Alerts related to the account.
    alert_timeline : LayoutDOM
        Timeline of alerts.
    related_bookmarks : pd.DataFrame
        Investigation bookmarks related to the account.
    host_logons : pd.DataFrame
        Host logon attemtps for selected account.
    host_logon_summary : pd.DataFrame
        Host logon summary for selected account.
    azure_activity : pd.DataFrame
        Azure Account activity for selected account.
    account_activity_summary : pd.DataFrame
        Azure activity summary.
    azure_timeline_by_provider : LayoutDOM
        Azure activity timeline grouped by provider
    account_timeline_by_ip : LayoutDOM
        Host or Azure activity timeline by IP Address.
    azure_timeline_by_operation : LayoutDOM
        Azure activity timeline grouped by operation
    ip_address_summary : pd.DataFrame
        Summary of IP address properties and usage for
        the current activity.
    ip_all_data : pd.DataFrame
        Full details of operations with IP WhoIs and GeoIP
        data.

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[, optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.description: str = "Account Activity Summary"
        self.account_entity: entities.Account = None
        self.account_activity: Optional[pd.DataFrame] = None
        self.account_selector: Optional[nbwidgets.SelectItem] = None
        self.related_alerts: Optional[pd.DataFrame] = None
        self.alert_timeline: Optional[LayoutDOM] = None
        self.related_bookmarks: Optional[pd.DataFrame] = None
        self.host_logons: Optional[pd.DataFrame] = None
        self.host_logon_summary: Optional[pd.DataFrame] = None
        self.azure_activity: Optional[pd.DataFrame] = None
        self.azure_activity_summary: Optional[pd.DataFrame] = None
        self.azure_timeline_by_provider: Optional[LayoutDOM] = None
        self.account_timeline_by_ip: Optional[LayoutDOM] = None
        self.azure_timeline_by_operation: Optional[LayoutDOM] = None
        self.ip_summary: Optional[pd.DataFrame] = None
        self.ip_all_data: Optional[pd.DataFrame] = None


# pylint: enable=too-few-public-methods


class AccountSummary(Notebooklet):
    """
    Retrieves account summary for the selected account.

    Main operations:

    - Searches for matches for the account name in Active Directory,
      Windows and Linux host logs.
    - If one or more matches are found it will return a selection
      widget that you can use to pick the account.
    - Selecting the account displays a summary of recent activity and
      retrieves any alerts and hunting bookmarks related to the account
    - The alerts and bookmarks are browseable using the `browse_alerts`
      and `browse_bookmarks` methods
    - You can call the `get_additional_data` method to retrieve and
      display more detailed activity information for the account.

    All of the returned data items are stored in the results class
    as entities, pandas DataFrames or Bokeh visualizations.
    Run help(nblt) on the notebooklet class to see usage.
    Run help(result) on the result class to see documentation of its
    properties.
    Run the print_options() method on either the notebooklet or
    results class to see information about the `options` parameter
    for the run() method.

    """

    # assign metadata from YAML to class variable
    metadata = _CLS_METADATA
    __doc__ = nb_metadata.update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    ACCOUNT_TYPE = AccountType

    def __init__(self, *args, **kwargs):
        """Initialize the Account Summary notebooklet."""
        super().__init__(*args, **kwargs)
        self._geo_lookup = None

    # pylint: disable=too-many-branches

    @set_text(docs=_CELL_DOCS, key="run")
    def run(
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> AccountSummaryResult:
        """
        Return account activity summary.

        Parameters
        ----------
        value : str
            Account name to search for.
        data : Optional[pd.DataFrame], optional
            Not used.
        timespan : TimeSpan
            Timespan for queries
        options : Optional[Iterable[str]], optional
            List of options to use, by default None.
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.
        account_types : Iterable[AccountType], Optional
            A list of account types to search for, by default
            all types.

        Returns
        -------
        AccountSummaryResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        # This line use logic in the superclass to populate options
        # (including default options) into this class.
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not value:
            raise MsticnbMissingParameterError("value")
        if not timespan:
            raise MsticnbMissingParameterError("timespan.")

        # Create a result class
        result = AccountSummaryResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )
        self.timespan = timespan
        acc_types = kwargs.get("account_types")
        if acc_types is None:
            acc_types = AccountType.__members__

        # Search for events for specified account types
        all_acct_dfs = _get_matching_accounts(
            self.query_provider,
            account=value,
            timespan=timespan,
            account_types=acc_types,
        )

        self._geo_lookup = self.get_provider("geolitelookup") or self.get_provider(
            "ipstacklookup"
        )

        acct_index_df = _create_account_index(all_acct_dfs)
        if acct_index_df.empty:
            nb_markdown("No accounts matching that name.")
        else:
            result.account_selector = _get_account_selector(
                qry_prov=self.query_provider,
                all_acct_dfs=all_acct_dfs,
                result=result,
                timespan=timespan,
                options=self.options,
                geoip=self._geo_lookup,
            )
        if len(acct_index_df) == 1:
            # If a single account - just display directly.
            disp_func = _create_display_callback(
                qry_prov=self.query_provider,
                all_acct_dfs=all_acct_dfs,
                result=result,
                timespan=timespan,
                options=self.options,
                geoip=self._geo_lookup,
            )
            acct_row = acct_index_df.iloc[0]
            outputs = disp_func(f"{acct_row.AccountName} {acct_row.Source}")
            for output in outputs:
                nb_display(output)
            if not self.silent:
                self.display_alert_timeline()
        elif len(acct_index_df) > 1:
            # if multiple, create a selector
            nb_markdown("<hr>")
            nb_markdown(
                "Multiple matching accounts found, select one to see details.", "large"
            )
            nb_display(result.account_selector)

        if not acct_index_df.empty:
            nb_markdown("<hr>")
            nb_markdown(
                "<h3>Use <i>result.notebooklet.get_additional_data()</i>"
                + " to retrieve more data."
            )
            nb_markdown(
                f"Additional methods for this class:<br>{'<br>'.join(self.list_methods())}"
            )
        # Assign the result to the _last_result attribute
        # so that you can get to it without having to re-run the operation
        self._last_result = result  # pylint: disable=attribute-defined-outside-init

        return self._last_result

    # pylint: enable=too-many-branches

    def display_alert_timeline(self):
        """Display the alert timeline."""
        if self.check_valid_result_data("related_alerts"):
            if len(self._last_result.related_alerts) > 1:
                return _get_alerts_timeline(
                    self._last_result.related_alerts, silent=False
                )
            print("Cannot plot timeline with 0 or 1 event.")
        return None

    def browse_accounts(self) -> nbwidgets.SelectItem:
        """Return the accounts browser/viewer."""
        if self.check_valid_result_data("account_selector"):
            return self._last_result.account_selector
        return None

    def browse_alerts(self) -> nbwidgets.SelectAlert:
        """Return alert browser/viewer."""
        if self.check_valid_result_data("related_alerts"):
            return browse_alerts(self._last_result)
        return None

    def browse_bookmarks(self) -> nbwidgets.SelectItem:
        """Return bookmark browser/viewer."""
        if self.check_valid_result_data("related_bookmarks"):
            return _get_bookmark_select(self._last_result.related_bookmarks)
        return None

    def az_activity_timeline_by_provider(self):
        """Display Azure activity timeline by provider."""
        if self.check_valid_result_data("azure_activity"):
            return _plot_timeline_by_provider(self._last_result.azure_activity)
        return None

    def az_activity_timeline_by_ip(self):
        """Display Azure activity timeline by IP address."""
        if self.check_valid_result_data("azure_activity"):
            return _plot_timeline_by_ip(self._last_result.azure_activity)
        return None

    def az_activity_timeline_by_operation(self):
        """Display Azure activity timeline by operation."""
        if self.check_valid_result_data("azure_activity"):
            return _plot_timeline_by_operation(self._last_result.azure_activity)
        return None

    def show_ip_summary(self):
        """Display Azure activity timeline by operation."""
        if self.check_valid_result_data("ip_summary"):
            nb_display(self._last_result.ip_summary)

    def host_logon_timeline(self):
        """Display IP address summary."""
        if self.check_valid_result_data("host_logons"):
            _, source = self._get_selected_account()
            ip_col = (
                "SourceIP"
                if AccountType.parse(source) == AccountType.Linux
                else "IpAddress"
            )
            return _create_host_timeline(
                self._last_result.host_logons, ip_col=ip_col, silent=False
            )
        return None

    def get_geoip_map(self):
        """Return Folium map of IP activity."""
        if self.check_valid_result_data("azure_activity", silent=True):
            return map_ips(
                data=self._last_result.azure_activity,
                ip_col="IPAddress",
                summary_cols=["SourceSystem", "Operation"],
                geo_lookup=self._geo_lookup,
            )
        if self.check_valid_result_data("host_logons", silent=True):
            ip_col = (
                "SourceIP"
                if "SourceIP" in self._last_result.host_logons.columns
                else "IpAddress"
            )
            return map_ips(
                data=self._last_result.host_logons,
                ip_col=ip_col,
                summary_cols=["Source", "Computer", "Operation"],
                geo_lookup=self._geo_lookup,
            )

        print(
            "Please ensure that you have run the 'run()'",
            "and 'get_additional_data()' methods before calling this.",
        )
        return None

    @set_text(docs=_CELL_DOCS, key="find_additional_data")
    def get_additional_data(self):
        """Find additional data for the selected account."""
        if not self.check_valid_result_data():
            return None
        acct, source = self._get_selected_account()
        if not acct or not source:
            print("Please use select an account before using this method.")
            return None
        self._last_result.host_logons = None
        self._last_result.host_logon_summary = None
        self._last_result.account_timeline_by_ip = None
        self._last_result.azure_activity = None
        self._last_result.azure_timeline_by_provider = None
        self._last_result.account_timeline_by_ip = None
        self._last_result.azure_timeline_by_operation = None
        self._last_result.azure_activity_summary = None
        self._last_result.ip_summary = None

        acct_type = AccountType.parse(source)
        if acct_type == AccountType.Linux:
            self._last_result.host_logons = _get_linux_add_activity(
                self.query_provider, acct, self.timespan
            )
            self._last_result.host_logon_summary = _summarize_host_activity(
                self._last_result.host_logons, ip_col="SourceIP"
            )
            self._last_result.account_timeline_by_ip = _create_host_timeline(
                self._last_result.host_logons, ip_col="SourceIP", silent=self.silent
            )
            (
                self._last_result.ip_summary,
                self._last_result.ip_all_data,
            ) = _create_ip_summary(
                self._last_result.host_logons,
                ip_col="SourceIP",
                geoip=self._geo_lookup,
            )
            nb_display(self._last_result.ip_summary)
            return None
        if acct_type == AccountType.Windows:
            self._last_result.host_logons = _get_windows_add_activity(
                self.query_provider, acct, self.timespan
            )
            self._last_result.host_logon_summary = _summarize_host_activity(
                self._last_result.host_logons, ip_col="IpAddress"
            )
            self._last_result.account_timeline_by_ip = _create_host_timeline(
                self._last_result.host_logons, ip_col="IpAddress", silent=self.silent
            )
            (
                self._last_result.ip_summary,
                self._last_result.ip_all_data,
            ) = _create_ip_summary(
                self._last_result.host_logons,
                ip_col="IpAddress",
                geoip=self._geo_lookup,
            )
            nb_display(self._last_result.ip_summary)
            return None
        if acct_type in [
            AccountType.AzureActiveDirectory,
            AccountType.AzureActivity,
            AccountType.Office365,
        ]:
            az_activity = _get_azure_add_activity(
                self.query_provider, acct, self.timespan
            )
            self._last_result.azure_activity = az_activity
            timelines = _create_azure_timelines(az_activity, self.silent)
            self._last_result.azure_timeline_by_provider = timelines[0]
            self._last_result.account_timeline_by_ip = timelines[1]
            self._last_result.azure_timeline_by_operation = timelines[2]
            self._last_result.azure_activity_summary = _summarize_azure_activity(
                az_activity
            )
            (
                self._last_result.ip_summary,
                self._last_result.ip_all_data,
            ) = _create_ip_summary(
                self._last_result.azure_activity,
                ip_col="IPAddress",
                geoip=self._geo_lookup,
            )
            nb_display(self._last_result.ip_summary)
            return None
        return None

    def _get_selected_account(self):
        if (
            self._last_result is not None
            and self._last_result.account_selector is not None
        ):
            return self._last_result.account_selector.value.split(" ")
        return "", ""


# pylint: disable=no-member


# %%
# Account Query functions
def _df_clean(dataframe):
    """Clean empty aggregate rows because of arg_max."""
    if isinstance(dataframe, pd.DataFrame):
        return dataframe[dataframe["TimeGenerated"].notna()]
    return pd.DataFrame()


@set_text(docs=_CELL_DOCS, key="get_matching_accounts")
def _get_matching_accounts(qry_prov, timespan, account, account_types):
    """Get Account Activity for `account` in `timespan`."""
    account_dfs = {}
    rec_count = 0
    if AccountType.AzureActiveDirectory.in_list(account_types):
        nb_data_wait("AADSignin")
        summarize_clause = f"""
        | summarize arg_max(TimeGenerated, *)
        | extend AccountName = UserPrincipalName,
          Source = '{AccountType.AzureActiveDirectory.name}'
        """

        aad_signin_df = qry_prov.Azure.list_aad_signins_for_account(
            timespan, account_name=account, add_query_items=summarize_clause
        )
        aad_signin_df = _df_clean(aad_signin_df)
        rec_count += len(aad_signin_df)
        nb_markdown(f"  {len(aad_signin_df)} records in AAD")
        account_dfs[AccountType.AzureActiveDirectory] = aad_signin_df

    if AccountType.Office365.in_list(account_types):
        nb_data_wait("Office365Activity")
        # Office Activity
        summarize_clause = f"""
        | extend UserId = tolower(UserId)
        | summarize arg_max(TimeGenerated, *)
        | extend IPAddress = case(
            isnotempty(Client_IPAddress), Client_IPAddress,
            isnotempty(ClientIP), ClientIP,
            isnotempty(ClientIP_), ClientIP_,
            Client_IPAddress
        )
        | extend AccountName = UserId,
          UserPrincipalName = UserId,
          Source = '{AccountType.Office365.name}'
        """

        o365_activity_df = qry_prov.Office365.list_activity_for_account(
            timespan, account_name=account, add_query_items=summarize_clause
        )
        o365_activity_df = _df_clean(o365_activity_df)
        rec_count += len(o365_activity_df)
        nb_markdown(f"  {len(o365_activity_df)} records in Office Activity")
        account_dfs[AccountType.Office365] = o365_activity_df

    if AccountType.Windows.in_list(account_types):
        nb_data_wait("Windows Logon activity")
        # Windows Host
        summarize_clause = f"""
        | extend LogonStatus = iff(EventID == 4624, "success", "failed")
        | summarize arg_max(TimeGenerated, *)
        | extend AccountName = TargetUserName,
          Source = '{AccountType.Windows.name}'
        """

        win_logon_df = qry_prov.WindowsSecurity.list_logon_attempts_by_account(
            timespan, account_name=account, add_query_items=summarize_clause
        )
        win_logon_df = _df_clean(win_logon_df)
        rec_count += len(win_logon_df)
        nb_markdown(f"  {len(win_logon_df)} records in Windows logon data")
        account_dfs[AccountType.Windows] = win_logon_df

    if AccountType.Linux.in_list(account_types):
        nb_data_wait("Linux logon activity")
        # Linux host
        summarize_clause = f"""
        | summarize arg_max(TimeGenerated, *)
        | extend Source = '{AccountType.Linux.name}'
        """

        linux_logon_df = qry_prov.LinuxSyslog.list_logons_for_account(
            timespan, account_name=account, add_query_items=summarize_clause
        )
        linux_logon_df = _df_clean(linux_logon_df)
        rec_count += len(linux_logon_df)
        nb_markdown(f"  {len(linux_logon_df)} records in Linux logon data")
        account_dfs[AccountType.Linux] = linux_logon_df

    nb_markdown(f"Found {rec_count} total records.")

    return account_dfs


# %%
# Account selector functions
def _display_acct_activity(
    selected_item: str, acct_activity_df: pd.DataFrame
) -> Iterable[Any]:
    """Return list of display objects for Select list display."""
    acct, source = selected_item.split(" ")
    outputs = []
    title = HTML(f"<b>{acct} (source: {source})</b>")
    outputs.append(title)
    outputs.append(acct_activity_df.sort_values("TimeGenerated", ascending=True))
    return outputs


def _get_select_acct_dict(acct_index_df: pd.DataFrame) -> Dict[str, str]:
    """Return dictionary of account entries for Select list."""

    def format_tuple(row):
        return (
            row.AccountName
            + "   "
            + row.Source
            + " (Last activity: "
            + str(row.TimeGenerated)
            + ")",
            row.AccountName + " " + row.Source,
        )

    return {item[0]: item[1] for item in acct_index_df.apply(format_tuple, axis=1)}


def _create_account_index(all_acct_dfs):
    acct_activity_df = pd.concat(all_acct_dfs.values())
    if acct_activity_df.empty:
        return acct_activity_df[["AccountName", "Source", "TimeGenerated"]]
    return (
        acct_activity_df[["AccountName", "Source", "TimeGenerated"]]
        .groupby(["AccountName", "Source"])
        .max()
        .reset_index()
        .dropna()
    )


def _get_account_selector(
    qry_prov, all_acct_dfs: pd.DataFrame, result, timespan, options, geoip
):
    """Build and return the Account Select list."""
    action_func = _create_display_callback(
        qry_prov=qry_prov,
        all_acct_dfs=all_acct_dfs,
        result=result,
        timespan=timespan,
        options=options,
        geoip=geoip,
    )

    acct_index_df = _create_account_index(all_acct_dfs)
    accts_dict = _get_select_acct_dict(acct_index_df)
    return nbwidgets.SelectItem(
        item_dict=accts_dict,
        description="Select an account to explore",
        action=action_func,
        height="200px",
        width="100%",
    )


def _create_display_callback(
    qry_prov, all_acct_dfs, result, timespan, options, geoip
) -> Callable[[str], Iterable[Any]]:
    """Create closure for display_acct callback."""

    def display_account(selected_item: str):
        account_name, source = selected_item.split(" ")
        acct_type = AccountType.parse(source)
        outputs = []

        # Create entity
        acct_entity = _create_account_entity(
            account_name, acct_type, all_acct_dfs, geoip
        )
        outputs.append(HTML("<h3>Account Entity</h3>"))
        result.account_entity = acct_entity
        outputs.append(acct_entity)
        # Add account activity
        outputs.append(HTML("<h3>Account last activity</h3>"))
        outputs.extend(_display_acct_activity(selected_item, all_acct_dfs[acct_type]))
        result.account_activity = all_acct_dfs[acct_type]

        # Optional alerts and bookmarks
        if "get_alerts" in options:
            related_alerts = _get_related_alerts(qry_prov, account_name, timespan)
            result.related_alerts = related_alerts
            outputs.append(_get_related_alerts_summary(related_alerts))
            if df_has_data(related_alerts):
                result.alert_timeline = _get_alerts_timeline(related_alerts)
        if "get_bookmarks" in options:
            related_bkmarks = _get_related_bookmarks(qry_prov, account_name, timespan)
            result.related_bookmarks = related_bkmarks
            outputs.append(_get_related_bkmks_summary(related_bkmarks))
        return outputs

    return display_account


# %%
# Account entity from data
def _create_account_entity(
    account_name, acct_type, acct_activity_dfs, geoip
) -> entities.Account:
    if acct_type == AccountType.Windows:
        acct_activity_df = acct_activity_dfs[AccountType.Windows]
        return _create_win_account_entity(account_name, acct_activity_df, geoip)

    if acct_type == AccountType.Linux:
        acct_activity_df = acct_activity_dfs[AccountType.Linux]
        return _create_lx_account_entity(account_name, acct_activity_df, geoip)

    if acct_type == AccountType.AzureActiveDirectory:
        acct_activity_df = acct_activity_dfs[AccountType.AzureActiveDirectory]
        return _create_aad_account_entity(account_name, acct_activity_df, geoip)

    if acct_type == AccountType.Office365:
        acct_activity_df = acct_activity_dfs[AccountType.Office365]
        return _create_o365_account_entity(account_name, acct_activity_df, geoip)

    acc_entity = entities.Account()
    acc_entity.Name = account_name
    return acc_entity


def _create_win_account_entity(
    account_name: str, acct_activity_df: pd.DataFrame, geoip
) -> entities.Account:
    account_events = acct_activity_df[acct_activity_df["AccountName"] == account_name]
    account_event = account_events.iloc[0]
    acc_entity = entities.Account(src_event=account_event)
    acc_entity.LogonType = account_event["LogonTypeName"]
    acc_entity.AadTenantId = account_event["TenantId"]

    host_grp = _create_host_ip_group(
        account_events, host_column="Computer", ip_column="IpAddress"
    )
    acc_hosts = list(_create_host_entities(host_grp, geoip))
    if len(acc_hosts) == 1:
        acc_entity.Host = acc_hosts[0]
    else:
        acc_entity.Hosts = acc_hosts
    return acc_entity


def _create_lx_account_entity(
    account_name: str, acct_activity_df: pd.DataFrame, geoip
) -> entities.Account:
    acc_entity = entities.Account()
    account_events = acct_activity_df[acct_activity_df["AccountName"] == account_name]
    account_event = account_events.iloc[0]
    acc_entity.Name = account_event["AccountName"]
    acc_entity.LogonType = account_event["LogonType"]
    acc_entity.Sid = account_event["UID"]
    acc_entity.AadTenantId = account_event["TenantId"]
    host_grp = _create_host_ip_group(
        account_events, host_column="Computer", ip_column="SourceIP"
    )
    acc_hosts = list(_create_host_entities(host_grp, geoip))
    if len(acc_hosts) == 1:
        acc_entity.Host = acc_hosts[0]
    else:
        acc_entity.Hosts = acc_hosts
    return acc_entity


def _create_host_ip_group(data, host_column, ip_column):
    """Group data by Host."""
    return data.groupby([host_column, ip_column]).agg(
        FirstSeen=pd.NamedAgg(column="TimeGenerated", aggfunc="min"),
        LastSeen=pd.NamedAgg(column="TimeGenerated", aggfunc="max"),
    )


# pylint: disable=too-many-branches
def _create_host_entities(data, geoip):
    """Create Host and IP Entities with GeoIP info."""
    for row in data.itertuples():
        host, ip_addr = row.Index
        if not host:
            continue
        host_ent = entities.Host(HostName=host)
        if not ip_addr:
            yield host_ent
            continue
        # If we have an IP address - get the IpAddress entity(ies)
        _, ip_entities = geoip.lookup_ip(ip_addr) if geoip else (None, [])
        if not ip_entities:
            host_ent.IpAddress = entities.IpAddress(Address=ip_addr)
            host_ent.IpAddress.FirstSeen = row.FirstSeen
            host_ent.IpAddress.LastSeen = row.LastSeen
        elif len(ip_entities) == 1:
            host_ent.IpAddress = ip_entities[0]
            host_ent.IpAddress.FirstSeen = row.FirstSeen
            host_ent.IpAddress.LastSeen = row.LastSeen
        else:
            host_ent.IpAddresses = []
            for ip_ent in ip_entities:
                ip_ent.FirstSeen = row.FirstSeen
                ip_ent.LastSeen = row.LastSeen
                host_ent.IpAddresses.append(ip_ent)

        yield host_ent


# pylint: enable=too-many-branches


def _create_aad_account_entity(
    account_name: str, acct_activity_df: pd.DataFrame, geoip
) -> entities.Account:
    acc_entity = entities.Account()
    account_events = acct_activity_df[
        (acct_activity_df["AccountName"] == account_name)
        & (acct_activity_df["Source"] == AccountType.AzureActiveDirectory.name)
    ]
    account_event = account_events.iloc[0]
    acc_entity.Name = account_event["UserPrincipalName"]
    if "@" in account_event["UserPrincipalName"]:
        acc_entity.UPNSuffix = account_event["UserPrincipalName"].split("@")[1]
    acc_entity.AadTenantId = account_event["AADTenantId"]
    acc_entity.AadUserId = account_event["UserId"]
    acc_entity.DisplayName = account_event["UserDisplayName"]
    acc_entity.DeviceDetail = account_event["DeviceDetail"]
    acc_entity.Location = account_event["LocationDetails"]
    acc_entity.UserAgent = account_event["UserAgent"]

    ip_grp = _create_ip_group(account_events, "IPAddress")
    ip_addrs = list(_create_ip_entities(ip_grp, geoip))
    if len(ip_addrs) == 1:
        acc_entity.IpAddress = ip_addrs[0]
    else:
        acc_entity.IpAddresses = ip_addrs
    return acc_entity


def _create_o365_account_entity(account_name, acct_activity_df, geoip):
    acc_entity = entities.Account()
    o365_events = acct_activity_df[
        (acct_activity_df["AccountName"] == account_name)
        & (acct_activity_df["Source"] == AccountType.Office365.name)
    ]
    account_event = o365_events.iloc[0]
    acc_entity.Name = account_event["UserId"]
    if "@" in account_event["UserId"]:
        acc_entity.UPNSuffix = account_event["UserId"].split("@")[1]
    acc_entity.AadTenantId = account_event["TenantId"]
    acc_entity.OrganizationId = account_event["OrganizationId"]

    ip_grp = _create_ip_group(o365_events, "IPAddress")
    ip_addrs = list(_create_ip_entities(ip_grp, geoip))
    if len(ip_addrs) == 1:
        acc_entity.IpAddress = ip_addrs[0]
    else:
        acc_entity.IpAddresses = ip_addrs

    return acc_entity


def _create_ip_group(data, ip_column):
    """Group data by IP Address."""
    return data.groupby(ip_column).agg(
        FirstSeen=pd.NamedAgg(column="TimeGenerated", aggfunc="min"),
        LastSeen=pd.NamedAgg(column="TimeGenerated", aggfunc="max"),
    )


def _create_ip_entities(data, geoip):
    """Create IP Entities with GeoIP info."""
    for row in data.itertuples():
        if row.Index:
            if not geoip:
                ip_ent = entities.IpAddress(Address=row.Index)
                ip_ent.FirstSeen = row.FirstSeen
                ip_ent.LastSeen = row.LastSeen
                yield ip_ent
            else:
                _, ip_entities = geoip.lookup_ip(row.Index)
                for ip_ent in ip_entities:
                    ip_ent.FirstSeen = row.FirstSeen
                    ip_ent.LastSeen = row.LastSeen
                    yield ip_ent


# %%
# Alert and bookmark details
def _get_related_alerts(
    qry_prov, account_name: str, timespan: TimeSpan
) -> pd.DataFrame:
    nb_data_wait("Alerts")
    return qry_prov.SecurityAlert.list_related_alerts(
        timespan, account_name=account_name
    )


def _get_alerts_timeline(related_alerts: pd.DataFrame, silent=True) -> LayoutDOM:
    """Return alert timeline."""
    return display_timeline(
        data=related_alerts,
        title="Alerts",
        source_columns=["AlertName"],
        height=300,
        hide=silent,
    )


def _get_related_alerts_summary(related_alerts: pd.DataFrame):
    if related_alerts.empty:
        return HTML("<br><b>No alerts for this account</b>")
    alert_items = (
        related_alerts[["AlertName", "TimeGenerated"]]
        .groupby("AlertName")
        .TimeGenerated.agg("count")
        .to_dict()
    )
    output = [
        "<h3>Related alerts</h3>",
        f"<b>Found {len(alert_items)} different alert types "
        f"related to this account</b>",
    ]

    total_alerts = 0
    for name, count in alert_items.items():
        output.append(f"- {name}, # Alerts: {count}")
        total_alerts += count

    if total_alerts > 1:
        output.append(
            "<br>To show the alert timeline call the <b>display_alert_timeline()</b> method."
        )
    output.append("To browse the alerts call the <b>browse_alerts()</b> method.")
    return HTML("<br>".join(output))


def _get_related_bookmarks(
    qry_prov, account_name: str, timespan: TimeSpan
) -> pd.DataFrame:
    nb_data_wait("Bookmarks")
    return qry_prov.AzureSentinel.list_bookmarks_for_entity(
        timespan, entity_id=account_name
    )


def _get_related_bkmks_summary(related_bookmarks: pd.DataFrame):
    if related_bookmarks.empty:
        return HTML("<br><b>No bookmarks for this account</b>")
    bookmarks = list(
        related_bookmarks.apply(
            lambda x: (
                f"{x.BookmarkName} [{x.Tags}]  {x.TimeGenerated}"
                + ", BookmarkId:{x.BookmarkId}"
            ),
            axis=1,
        )
    )
    output = [
        "<h3>Related bookmarks</h3>",
        f"<b>Found {len(bookmarks)} different bookmarks " "related to this account</b>",
        *bookmarks,
        "<br>To browse the bookmarks call the <b>browse_bookmarks()</b> method.",
    ]

    return HTML("<br>".join(output))


def _get_bookmark_select(bookmarks_df):
    """Create and return Selector for bookmarks."""
    opts = dict(
        bookmarks_df.apply(
            lambda x: (
                f"{x.BookmarkName} - LastUpdated {x.LastUpdatedTime}",
                x.BookmarkId,
            ),
            axis=1,
        ).values
    )

    def display_bookmark(bookmark_id):
        return pd.DataFrame(
            data=bookmarks_df[bookmarks_df["BookmarkId"] == bookmark_id].iloc[0].T
        ).rename(columns={0: "value"})

    return nbwidgets.SelectItem(
        item_dict=opts, action=display_bookmark, height="200px", width="100%"
    )


# %%
# Get Linux logon activity
def _get_linux_add_activity(qry_prov, acct, timespan):
    nb_data_wait("LinuxSyslog")
    ext_logon_result = "| extend Operation = strcat('Logon-', LogonResult)"
    return qry_prov.LinuxSyslog.list_logons_for_account(
        timespan, account_name=acct, add_query_items=ext_logon_result
    )


@set_text(docs=_CELL_DOCS, key="summarize_host_activity")
def _summarize_host_activity(all_logons: pd.DataFrame, ip_col="SourceIP"):
    """Summarize logon activity on win or linux host."""
    summary = all_logons.groupby("Computer").agg(
        TotalLogons=pd.NamedAgg(column="Computer", aggfunc="count"),
        FailedLogons=pd.NamedAgg(
            column="LogonResult", aggfunc=lambda x: x.value_counts().to_dict()
        ),
        IpAddresses=pd.NamedAgg(column=ip_col, aggfunc=lambda x: x.unique().tolist()),
        LogonTypeCount=pd.NamedAgg(
            column="LogonType", aggfunc=lambda x: x.value_counts().to_dict()
        ),
        FirstLogon=pd.NamedAgg(column="TimeGenerated", aggfunc="min"),
        LastLogon=pd.NamedAgg(column="TimeGenerated", aggfunc="max"),
    )
    nb_display(summary)
    return summary


@set_text(docs=_CELL_DOCS, key="create_host_timeline")
def _create_host_timeline(
    all_logons: pd.DataFrame, ip_col="SourceIP", silent: bool = False
):
    return display_timeline(
        data=all_logons,
        group_by=ip_col,
        source_columns=["Computer", "LogonResult", "LogonType"],
        title="Timeline of Logons by source IP address",
        hide=silent,
    )


# %%
# Get Windows logon activity
def _get_windows_add_activity(qry_prov, acct, timespan):
    nb_data_wait("WindowsSecurity")
    ext_logon_result = """
        | extend LogonResult = iff(EventID == 4624, 'Success', 'Failed')
        | extend Operation = strcat("Logon-", LogonResult)
    """
    return qry_prov.WindowsSecurity.list_logon_attempts_by_account(
        timespan, account_name=acct, add_query_items=ext_logon_result
    )


# %%
# Get Azure/AAD/Office activity
def _get_azure_add_activity(qry_prov, acct, timespan):
    """Get Azure additional data for account."""
    nb_data_wait("AADSignin")
    aad_sum_qry = f"""
        | extend UserPrincipalName=tolower(UserPrincipalName)
        | extend Source = '{AccountType.AzureActiveDirectory.name}'
        | project-rename Operation=OperationName, AppResourceProvider=AppDisplayName
    """
    aad_logons = qry_prov.Azure.list_aad_signins_for_account(
        timespan, account_name=acct, add_query_items=aad_sum_qry
    )

    nb_data_wait("AzureActivity")
    az_sum_qry = f"""
        | extend UserPrincipalName=tolower(Caller)
        | extend Source = '{AccountType.AzureActivity.name}'
        | project-rename IPAddress=CallerIpAddress, Operation=OperationName,
        AppResourceProvider=ResourceProvider
    """
    az_activity = qry_prov.Azure.list_azure_activity_for_account(
        timespan, account_name=acct, add_query_items=az_sum_qry
    )

    nb_data_wait("Office365Activity")
    o365_sum_qry = f"""
        | extend UserPrincipalName=tolower(UserId)
        | extend Source = '{AccountType.Office365.name}'
        | extend IPAddress = case(
            isnotempty(Client_IPAddress), Client_IPAddress,
            isnotempty(ClientIP), ClientIP,
            isnotempty(ClientIP_), ClientIP_,
            Client_IPAddress
        )
        | project-rename ResourceId=OfficeObjectId, AppResourceProvider=OfficeWorkload
    """
    o365_activity = qry_prov.Office365.list_activity_for_account(
        timespan, account_name=acct, add_query_items=o365_sum_qry
    )

    return pd.concat(
        [aad_logons, az_activity, o365_activity], ignore_index=True, sort=False
    ).sort_values("TimeGenerated")


@set_text(docs=_CELL_DOCS, key="create_az_timelines")
def _create_azure_timelines(az_all_data: pd.DataFrame, silent: bool = False):
    return (
        _plot_timeline_by_provider(az_all_data, silent),
        _plot_timeline_by_ip(az_all_data, silent),
        _plot_timeline_by_operation(az_all_data, silent),
    )


def _plot_timeline_by_provider(az_all_data, silent=False):
    return display_timeline(
        data=az_all_data,
        group_by="AppResourceProvider",
        source_columns=["Operation", "IPAddress", "AppResourceProvider"],
        title="Azure account activity by Provider",
        hide=silent,
    )


def _plot_timeline_by_ip(az_all_data, silent=False):
    return display_timeline(
        data=az_all_data,
        group_by="IPAddress",
        source_columns=[
            "AppResourceProvider",
            "Operation",
            "IPAddress",
            "AppResourceProvider",
        ],
        title="Azure Operations by Source IP",
        hide=silent,
    )


def _plot_timeline_by_operation(az_all_data, silent=False):
    return display_timeline(
        data=az_all_data,
        group_by="Operation",
        source_columns=["AppResourceProvider", "Operation", "IPAddress"],
        title="Azure Operations by Operation",
        hide=silent,
    )


@set_text(docs=_CELL_DOCS, key="summarize_azure_activity")
def _summarize_azure_activity(az_all_data: pd.DataFrame):
    summary = az_all_data.groupby(
        ["UserPrincipalName", "Type", "IPAddress", "AppResourceProvider", "UserType"]
    ).agg(
        OperationCount=pd.NamedAgg(column="Type", aggfunc="count"),
        OperationTypes=pd.NamedAgg(
            column="Operation", aggfunc=lambda x: x.unique().tolist()
        ),
        Resources=pd.NamedAgg(column="ResourceId", aggfunc="nunique"),
        FirstOperation=pd.NamedAgg(column="TimeGenerated", aggfunc="min"),
        LastOperation=pd.NamedAgg(column="TimeGenerated", aggfunc="max"),
    )
    nb_display(summary)
    return summary


@set_text(docs=_CELL_DOCS, key="get_ip_summary")
def _create_ip_summary(data, ip_col, geoip):
    group_cols1 = ["Source", "Operation"]
    group_cols2 = ["IpAddress", "AsnDescription", "CountryCode"]
    if "UserAgent" in data.columns:
        group_cols1.append("UserAgent")
    group_cols = group_cols1 + group_cols2
    all_data = (
        data[[ip_col]]  # the property and the column we want
        .dropna()
        .drop_duplicates()  # drop duplicates
        .pipe(
            (get_geoip_whois, "data"), geo_lookup=geoip, ip_col=ip_col
        )  # get geoip and whois
        .drop(columns=["TimeGenerated", "Type"], errors="ignore")
        .merge(data, left_on="IpAddress", right_on=ip_col)
    )
    for col in group_cols:
        if col not in all_data.columns:
            group_cols.pop(group_cols.index(col))
    ip_summary = all_data.groupby(group_cols).agg(
        Count=pd.NamedAgg("Operation", "count"),
        FirstOperation=pd.NamedAgg("TimeGenerated", "min"),
        LastOperation=pd.NamedAgg("TimeGenerated", "max"),
    )

    return ip_summary, all_data
